#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Tests for `opencue.wrappers.job`."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import getpass
import os
import platform
import unittest

import mock

from opencue_proto import comment_pb2
from opencue_proto import depend_pb2
from opencue_proto import job_pb2
import opencue.wrappers.frame
import opencue.wrappers.group
import opencue.wrappers.job
import opencue.wrappers.layer


TEST_JOB_NAME = 'testJob'


@mock.patch('opencue.cuebot.Cuebot.getStub')
class JobTests(unittest.TestCase):

    def testKill(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.Delete.return_value = job_pb2.JobKillResponse()
        getStubMock.return_value = stubMock

        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        username = getpass.getuser()
        pid = os.getpid()
        host_kill = platform.uname()[1]
        reason = "Job Kill Request"
        job.kill(username=username, pid=pid, host_kill=host_kill, reason=reason)

        stubMock.Kill.assert_called_with(
            job_pb2.JobKillRequest(job=job.data,
                                   username=username,
                                   pid=str(pid),
                                   host_kill=host_kill,
                                   reason=reason), timeout=mock.ANY)

    def testPause(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.Pause.return_value = job_pb2.JobPauseResponse()
        getStubMock.return_value = stubMock

        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.pause()

        stubMock.Pause.assert_called_with(
            job_pb2.JobPauseRequest(job=job.data), timeout=mock.ANY)

    def testResume(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.Resume.return_value = job_pb2.JobResumeResponse()
        getStubMock.return_value = stubMock

        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.resume()

        stubMock.Resume.assert_called_with(
            job_pb2.JobResumeRequest(job=job.data), timeout=mock.ANY)

    def testKillFrames(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.KillFrames.return_value = job_pb2.JobKillFramesResponse()
        getStubMock.return_value = stubMock

        frameRange = '1-10'
        criteria = opencue.search.FrameSearch.criteriaFromOptions(range=frameRange)
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        username = getpass.getuser()
        pid = os.getpid()
        host_kill = platform.uname()[1]
        reason = "Job Kill Request"
        job.killFrames(range=frameRange,
                       username=username,
                       pid=str(pid),
                       host_kill=host_kill,
                       reason=reason)

        stubMock.KillFrames.assert_called_with(
            job_pb2.JobKillFramesRequest(job=job.data,
                                         username=username,
                                         pid=str(pid),
                                         host_kill=host_kill,
                                         reason=reason,
                                         req=criteria), timeout=mock.ANY)

    def testEatFrames(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.EatFrames.return_value = job_pb2.JobEatFramesResponse()
        getStubMock.return_value = stubMock

        frameRange = '1-10'
        criteria = opencue.search.FrameSearch.criteriaFromOptions(range=frameRange)
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.eatFrames(range=frameRange)

        stubMock.EatFrames.assert_called_with(
            job_pb2.JobEatFramesRequest(job=job.data, req=criteria), timeout=mock.ANY)

    def testRetryFrames(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.RetryFrames.return_value = job_pb2.JobRetryFramesResponse()
        getStubMock.return_value = stubMock

        frameRange = '1-10'
        criteria = opencue.search.FrameSearch.criteriaFromOptions(range=frameRange)
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.retryFrames(range=frameRange)

        stubMock.RetryFrames.assert_called_with(
            job_pb2.JobRetryFramesRequest(job=job.data, req=criteria), timeout=mock.ANY)

    def testMarkdoneFrames(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.MarkDoneFrames.return_value = job_pb2.JobMarkDoneFramesResponse()
        getStubMock.return_value = stubMock

        frameRange = '1-10'
        criteria = opencue.search.FrameSearch.criteriaFromOptions(range=frameRange)
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.markdoneFrames(range=frameRange)

        stubMock.MarkDoneFrames.assert_called_with(
            job_pb2.JobMarkDoneFramesRequest(job=job.data, req=criteria), timeout=mock.ANY)

    def testMarkAsWaiting(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.MarkAsWaiting.return_value = job_pb2.JobMarkAsWaitingResponse()
        getStubMock.return_value = stubMock

        frameRange = '1-10'
        criteria = opencue.search.FrameSearch.criteriaFromOptions(range=frameRange)
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.markAsWaiting(range=frameRange)

        stubMock.MarkAsWaiting.assert_called_with(
            job_pb2.JobMarkAsWaitingRequest(job=job.data, req=criteria), timeout=mock.ANY)

    def testSetMinCores(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.SetMinCores.return_value = job_pb2.JobSetMinCoresResponse()
        getStubMock.return_value = stubMock

        cores = 2000
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.setMinCores(cores)

        stubMock.SetMinCores.assert_called_with(
            job_pb2.JobSetMinCoresRequest(job=job.data, val=cores), timeout=mock.ANY)

    def testSetMaxCores(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.SetMaxCores.return_value = job_pb2.JobSetMaxCoresResponse()
        getStubMock.return_value = stubMock

        cores = 2000
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.setMaxCores(cores)

        stubMock.SetMaxCores.assert_called_with(
            job_pb2.JobSetMaxCoresRequest(job=job.data, val=cores), timeout=mock.ANY)

    def testSetPriority(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.SetPriority.return_value = job_pb2.JobSetPriorityResponse()
        getStubMock.return_value = stubMock

        priority = 50
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.setPriority(priority)

        stubMock.SetPriority.assert_called_with(
            job_pb2.JobSetPriorityRequest(job=job.data, val=priority), timeout=mock.ANY)

    def testSetMaxRetries(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.SetMaxRetries.return_value = job_pb2.JobSetMaxRetriesResponse()
        getStubMock.return_value = stubMock

        retries = 10
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.setMaxRetries(retries)

        stubMock.SetMaxRetries.assert_called_with(
            job_pb2.JobSetMaxRetriesRequest(job=job.data, max_retries=retries), timeout=mock.ANY)

    def testGetLayers(self, getStubMock):
        layerNames = ['testLayerA', 'testLayerB']
        stubMock = mock.Mock()
        stubMock.GetLayers.return_value = job_pb2.JobGetLayersResponse(
            layers=job_pb2.LayerSeq(layers=[job_pb2.Layer(name=layerNames[0]),
                                            job_pb2.Layer(name=layerNames[1])]))
        getStubMock.return_value = stubMock

        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        layers = job.getLayers()

        stubMock.GetLayers.assert_called_with(
            job_pb2.JobGetLayersRequest(job=job.data), timeout=mock.ANY)
        self.assertTrue(len(layers), 2)
        self.assertTrue(layers[0].name(), layerNames[0])
        self.assertTrue(layers[1].name(), layerNames[1])

    def testGetFrames(self, getStubMock):
        frameNames = ['testFrameA', 'testFrameB']
        stubMock = mock.Mock()
        stubMock.GetFrames.return_value = job_pb2.JobGetFramesResponse(
            frames=job_pb2.FrameSeq(frames=[job_pb2.Frame(name=frameNames[0]),
                                            job_pb2.Frame(name=frameNames[1])]))
        getStubMock.return_value = stubMock

        frameRange = '1-10'
        criteria = opencue.search.FrameSearch.criteriaFromOptions(range=frameRange)
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        frames = job.getFrames(range=frameRange)

        stubMock.GetFrames.assert_called_with(
            job_pb2.JobGetFramesRequest(job=job.data, req=criteria), timeout=mock.ANY)
        self.assertTrue(len(frames), 2)
        self.assertTrue(frames[0].name(), frameNames[0])
        self.assertTrue(frames[1].name(), frameNames[1])

    def testGetUpdatedFrames(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.GetUpdatedFrames.return_value = job_pb2.JobGetUpdatedFramesResponse(
            state=job_pb2.FINISHED,
            server_time=987654321,
            updated_frames=job_pb2.UpdatedFrameSeq(
                updated_frames=[job_pb2.UpdatedFrame(id='uuu-uuuu-uuu')]))
        getStubMock.return_value = stubMock

        lastCheck = 123456789
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        framesResponse = job.getUpdatedFrames(lastCheck)

        stubMock.GetUpdatedFrames.assert_called_with(
            job_pb2.JobGetUpdatedFramesRequest(job=job.data, last_check=lastCheck,
                                               layer_filter=None),
            timeout=mock.ANY)
        self.assertEqual(framesResponse.state, job_pb2.FINISHED)
        self.assertEqual(len(framesResponse.updated_frames.updated_frames), 1)

    def testSetAutoEating(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.SetAutoEat.return_value = job_pb2.JobSetAutoEatResponse()
        getStubMock.return_value = stubMock

        value = True
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.setAutoEating(value)

        stubMock.SetAutoEat.assert_called_with(
            job_pb2.JobSetAutoEatRequest(job=job.data, value=value), timeout=mock.ANY)

    def testGetWhatDependsOnThis(self, getStubMock):
        dependId = 'ddd-dddd-ddd'
        stubMock = mock.Mock()
        stubMock.GetWhatDependsOnThis.return_value = job_pb2.JobGetWhatDependsOnThisResponse(
            depends=depend_pb2.DependSeq(depends=[depend_pb2.Depend(id=dependId)]))
        getStubMock.return_value = stubMock

        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        depends = job.getWhatDependsOnThis()

        stubMock.GetWhatDependsOnThis.assert_called_with(
            job_pb2.JobGetWhatDependsOnThisRequest(job=job.data), timeout=mock.ANY)
        self.assertEqual(len(depends), 1)
        self.assertEqual(depends[0].id(), dependId)

    def testGetWhatThisDependsOn(self, getStubMock):
        dependId = 'ddd-dddd-ddd'
        stubMock = mock.Mock()
        stubMock.GetWhatThisDependsOn.return_value = job_pb2.JobGetWhatThisDependsOnResponse(
            depends=depend_pb2.DependSeq(depends=[depend_pb2.Depend(id=dependId)]))
        getStubMock.return_value = stubMock

        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        depends = job.getWhatThisDependsOn()

        stubMock.GetWhatThisDependsOn.assert_called_with(
            job_pb2.JobGetWhatThisDependsOnRequest(job=job.data), timeout=mock.ANY)
        self.assertEqual(len(depends), 1)
        self.assertEqual(depends[0].id(), dependId)

    def testGetDepends(self, getStubMock):
        dependId = 'ddd-dddd-ddd'
        stubMock = mock.Mock()
        stubMock.GetDepends.return_value = job_pb2.JobGetDependsResponse(
            depends=depend_pb2.DependSeq(depends=[depend_pb2.Depend(id=dependId)]))
        getStubMock.return_value = stubMock

        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        depends = job.getDepends()

        stubMock.GetDepends.assert_called_with(
            job_pb2.JobGetDependsRequest(job=job.data),
            timeout=mock.ANY)
        self.assertEqual(len(depends), 1)
        self.assertEqual(depends[0].id(), dependId)

    def testDropDepends(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.DropDepends.return_value = job_pb2.JobDropDependsResponse()
        getStubMock.return_value = stubMock

        target = depend_pb2.EXTERNAL
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.dropDepends(target)

        stubMock.DropDepends.assert_called_with(
            job_pb2.JobDropDependsRequest(job=job.data, target=target),
            timeout=mock.ANY)

    def testCreateDependencyOnJob(self, getStubMock):
        dependId = 'ddd-dddd-ddd'
        stubMock = mock.Mock()
        stubMock.CreateDependencyOnJob.return_value = job_pb2.JobCreateDependencyOnJobResponse(
            depend=depend_pb2.Depend(id=dependId))
        getStubMock.return_value = stubMock

        onJob = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME+"Depend"))
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        depend = job.createDependencyOnJob(onJob)

        stubMock.CreateDependencyOnJob.assert_called_with(
            job_pb2.JobCreateDependencyOnJobRequest(job=job.data, on_job=onJob.data),
            timeout=mock.ANY)
        self.assertEqual(depend.id(), dependId)

    def testCreateDependencyOnLayer(self, getStubMock):
        dependId = 'ddd-dddd-ddd'
        dependLayer = 'testLayer'
        stubMock = mock.Mock()
        stubMock.CreateDependencyOnLayer.return_value = job_pb2.JobCreateDependencyOnLayerResponse(
            depend=depend_pb2.Depend(id=dependId))
        getStubMock.return_value = stubMock

        onLayer = opencue.wrappers.layer.Layer(
            job_pb2.Layer(name=dependLayer))
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        depend = job.createDependencyOnLayer(onLayer)

        stubMock.CreateDependencyOnLayer.assert_called_with(
            job_pb2.JobCreateDependencyOnLayerRequest(job=job.data, layer=onLayer.data),
            timeout=mock.ANY)
        self.assertEqual(depend.id(), dependId)

    def testCreateDependencyOnFrame(self, getStubMock):
        dependId = 'ddd-dddd-ddd'
        dependFrame = 'testFrame'
        stubMock = mock.Mock()
        stubMock.CreateDependencyOnFrame.return_value = job_pb2.JobCreateDependencyOnFrameResponse(
            depend=depend_pb2.Depend(id=dependId))
        getStubMock.return_value = stubMock

        onFrame = opencue.wrappers.frame.Frame(
            job_pb2.Frame(name=dependFrame))
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        depend = job.createDependencyOnFrame(onFrame)

        stubMock.CreateDependencyOnFrame.assert_called_with(
            job_pb2.JobCreateDependencyOnFrameRequest(job=job.data, frame=onFrame.data),
            timeout=mock.ANY)
        self.assertEqual(depend.id(), dependId)

    def testAddComment(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.AddComment.return_value = job_pb2.JobAddCommentResponse()
        getStubMock.return_value = stubMock

        subject = 'test'
        message = 'this is a test.'
        comment = comment_pb2.Comment(user=os.getenv('USER', 'unknown'), subject=subject,
                                      message=message, timestamp=0)
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.addComment(subject, message)

        stubMock.AddComment.assert_called_with(
            job_pb2.JobAddCommentRequest(job=job.data, new_comment=comment),
            timeout=mock.ANY)

    def testGetComments(self, getStubMock):
        message = 'this is a test.'
        stubMock = mock.Mock()
        stubMock.GetComments.return_value = job_pb2.JobGetCommentsResponse(
            comments=comment_pb2.CommentSeq(comments=[comment_pb2.Comment(message=message)]))
        getStubMock.return_value = stubMock

        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        comments = job.getComments()

        stubMock.GetComments.assert_called_with(
            job_pb2.JobGetCommentsRequest(job=job.data),
            timeout=mock.ANY)
        self.assertEqual(len(comments), 1)
        self.assertEqual(comments[0].message(), message)

    def testSetGroup(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.SetGroup.return_value = job_pb2.JobSetGroupResponse()
        getStubMock.return_value = stubMock

        groupId = 'ggg-gggg-ggg'
        group = opencue.wrappers.group.Group(
            job_pb2.Group(id=groupId))
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.setGroup(group)

        stubMock.SetGroup.assert_called_with(
            job_pb2.JobSetGroupRequest(job=job.data, group_id=groupId),
            timeout=mock.ANY)

    def testReorderFrames(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.ReorderFrames.return_value = job_pb2.JobReorderFramesResponse()
        getStubMock.return_value = stubMock

        frameRange = '1-10'
        order = job_pb2.REVERSE
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.reorderFrames(frameRange, order)

        stubMock.ReorderFrames.assert_called_with(
            job_pb2.JobReorderFramesRequest(job=job.data, range=frameRange, order=order),
            timeout=mock.ANY)

    def testStaggerFrames(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.StaggerFrames.return_value = job_pb2.JobStaggerFramesResponse()
        getStubMock.return_value = stubMock

        frameRange = '1-10'
        stagger = 5
        job = opencue.wrappers.job.Job(
            job_pb2.Job(name=TEST_JOB_NAME))
        job.staggerFrames(frameRange, stagger)

        stubMock.StaggerFrames.assert_called_with(
            job_pb2.JobStaggerFramesRequest(job=job.data, range=frameRange, stagger=stagger),
            timeout=mock.ANY)

    def testFrameStateTotals(self, getStubMock):
        runningFrames = 10
        waitingFrames = 50
        succeededFrames = 30
        expected = {job_pb2.WAITING: waitingFrames,
                    job_pb2.RUNNING: runningFrames,
                    job_pb2.SUCCEEDED: succeededFrames,
                    job_pb2.CHECKPOINT: 0,
                    job_pb2.SETUP: 0,
                    job_pb2.EATEN: 0,
                    job_pb2.DEAD: 0,
                    job_pb2.DEPEND: 0}

        job = opencue.wrappers.job.Job(job_pb2.Job(
            name="frameStateTestJob",
            job_stats=job_pb2.JobStats(
                running_frames=runningFrames,
                waiting_frames=waitingFrames,
                succeeded_frames=succeededFrames)))

        frameStateTotals = job.frameStateTotals()
        self.assertEqual(frameStateTotals, expected)


@mock.patch('opencue.cuebot.Cuebot.getStub')
class NestedJobTests(unittest.TestCase):

    def testAsJob(self, getStubMock):
        jobId = 'nnn-nnnn-nnn'
        state = job_pb2.FINISHED
        name = 'testNestedJob'
        shot = 'shotname'
        show = 'showname'
        user = 'username'
        group = 'groupname'
        facility = 'facilityname'
        jobOs = 'os'
        uid = 12345
        priority = 14
        minCores = 5
        maxCores = 6
        logDir = '/path/to/logs'
        isPaused = False
        nestedJob = opencue.wrappers.job.NestedJob(
            job_pb2.NestedJob(id=jobId, state=state, name=name, shot=shot, show=show, user=user,
                              group=group, facility=facility, os=jobOs, uid=uid, priority=priority,
                              min_cores=minCores, max_cores=maxCores, log_dir=logDir,
                              is_paused=isPaused))
        job = opencue.wrappers.job.Job(
            job_pb2.Job(id=jobId, state=state, name=name, shot=shot, show=show, user=user,
                        group=group, facility=facility, os=jobOs, uid=uid, priority=priority,
                        min_cores=minCores, max_cores=maxCores, log_dir=logDir,
                        is_paused=isPaused))

        asJob = nestedJob.asJob()
        attrs = ['id', 'state', 'name', 'shot', 'show', 'user', 'group', 'facility', 'os', 'uid',
                 'priority', 'minCores', 'maxCores', 'logDir', 'isPaused']
        for attr in attrs:
            self.assertEqual(getattr(job, attr)(), getattr(asJob, attr)())
            self.assertEqual(getattr(nestedJob, attr)(), getattr(asJob, attr)())


class JobEnumTests(unittest.TestCase):

    def testJobState(self):
        self.assertEqual(opencue.api.Job.JobState.PENDING, job_pb2.PENDING)
        self.assertEqual(opencue.api.Job.JobState.PENDING, 0)


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module for classes related to shows."""

from opencue_proto import show_pb2
from opencue.cuebot import Cuebot
import opencue.wrappers.filter
import opencue.wrappers.group
import opencue.wrappers.subscription
from opencue.wrappers.service import ServiceOverride


class Show(object):
    """This class contains the grpc implementation related to a Show."""

    def __init__(self, show=None):
        self.data = show
        self.stub = Cuebot.getStub('show')

    def createOwner(self, user):
        """Creates a new owner for the show.

        :type  user: str
        :param user: user name
        :rtype:  host_pb2.Owner
        :return: the created owner object
        """
        response = self.stub.CreateOwner(show_pb2.ShowCreateOwnerRequest(show=self.data, name=user),
                                         timeout=Cuebot.Timeout)
        return response.owner

    def createSubscription(self, allocation, size, burst):
        """Creates a new subscription for the show.

        A subscription links a show to an allocation, and determines how many cores the show
        can utilize within that allocation.

        :type  allocation: opencue.wrappers.allocation.Allocation
        :param allocation: allocation to subscribe to
        :type  size: float
        :param size: number of cores the show is allowed to use consistently
        :type  burst: float
        :param burst: number of cores the show is allowed to burst to
        :rtype:  opencue.wrappers.subscription.Subscription
        :return: the created subscription object
        """
        response = self.stub.CreateSubscription(show_pb2.ShowCreateSubscriptionRequest(
            show=self.data, allocation_id=allocation.id(), size=size, burst=burst),
            timeout=Cuebot.Timeout)
        return opencue.wrappers.subscription.Subscription(response.subscription)

    def delete(self):
        """Deletes this show."""
        self.stub.Delete(show_pb2.ShowDeleteRequest(show=self.data), timeout=Cuebot.Timeout)

    def archive(self, target_show_name):
        """Archives this show by creating an alias to another show.

        When a show is archived, the original show is renamed with '_archive' suffix
        and aliased to the target show. Jobs submitted to the archived show will run
        on the target show's allocations.

        :type  target_show_name: str
        :param target_show_name: name of the show to alias to
        """
        self.stub.Archive(
            show_pb2.ShowArchiveRequest(show=self.data, target_show_name=target_show_name),
            timeout=Cuebot.Timeout)

    def createServiceOverride(self, data):
        """Creates a Service Override at the show level.
        :type data: opencue.wrapper.service.Service
        :param data: Service.data object
        """

        # min_memory_increase has to be greater than 0.
        if data.min_memory_increase <= 0:
            raise ValueError("Minimum memory increase must be > 0")
        self.stub.CreateServiceOverride(show_pb2.ShowCreateServiceOverrideRequest(
                                        show=self.data, service=data),
                                        timeout=Cuebot.Timeout)

    def getServiceOverride(self, serviceName):
        """
        Returns a service override for a show

        :param serviceName: name of the service for the show
        :return: service override object
        """
        serviceOverride = self.stub.GetServiceOverride(show_pb2.ShowGetServiceOverrideRequest(
                                                       show=self.data, name=serviceName),
                                                       timeout=Cuebot.Timeout).service_override
        return ServiceOverride(serviceOverride)

    def getServiceOverrides(self):
        """Returns a list of service overrides on the show.

        :rtype:  list<service_pb2.ServiceOverride>
        :return: list of service overrides on the show
        """
        serviceOverrideSeq = self.stub.GetServiceOverrides(
            show_pb2.ShowGetServiceOverridesRequest(show=self.data),
            timeout=Cuebot.Timeout).service_overrides
        return [ServiceOverride(override) for override in serviceOverrideSeq.service_overrides]

    def getSubscriptions(self):
        """Returns a list of all subscriptions the show has.

        :rtype:  list<opencue.wrappers.subscription.Subscription>
        :return: list of the show's subscriptions
        """
        response = self.stub.GetSubscriptions(
            show_pb2.ShowGetSubscriptionRequest(show=self.data), timeout=Cuebot.Timeout)
        subscriptionSeq = response.subscriptions
        return [opencue.wrappers.subscription.Subscription(subs)
                for subs in subscriptionSeq.subscriptions]

    @staticmethod
    def findSubscription(name):
        """Returns the matching subscription.

        :type name: str
        :param name: name of subscription to find
        :rtype:  opencue.wrappers.subscription.Subscription
        :return: the matching subscription
        """
        subscriptions = opencue.wrappers.subscription.Subscription()
        return subscriptions.find(name)

    def getFilters(self):
        """Returns the job filters for this show.

        :rtype:  list<opencue.wrappers.filter.Filter>
        :return: list of filters for this show
        """
        response = self.stub.GetFilters(
            show_pb2.ShowGetFiltersRequest(show=self.data), timeout=Cuebot.Timeout)
        filterSeq = response.filters
        return [opencue.wrappers.filter.Filter(filter) for filter in filterSeq.filters]

    def setActive(self, value):
        """Sets whether this show is active.

        :type  value: bool
        :param value: whether the show is active
        """
        self.stub.SetActive(show_pb2.ShowSetActiveRequest(show=self.data, value=value),
                            timeout=Cuebot.Timeout)

    def setDefaultMaxCores(self, maxcores):
        """Sets the default maximum number of cores that new jobs are launched with.

        :type  maxcores: float
        :param maxcores: new maximum number of cores for new jobs
        :rtype:  show_pb2.ShowSetDefaultMaxCoresResponse
        :return: response is empty
        """
        response = self.stub.SetDefaultMaxCores(show_pb2.ShowSetDefaultMaxCoresRequest(
            show=self.data, max_cores=maxcores),
            timeout=Cuebot.Timeout)
        return response

    def setDefaultMinCores(self, mincores):
        """Sets the default minimum number of cores new jobs are launched with.

        :type  mincores: float
        :param mincores: new minimum number of cores for new jobs
        :rtype:  show_pb2.ShowSetDefaultMinCoresResponse
        :return: response is empty
        """
        response = self.stub.SetDefaultMinCores(show_pb2.ShowSetDefaultMinCoresRequest(
            show=self.data, min_cores=mincores),
            timeout=Cuebot.Timeout)
        return response

    def setDefaultMaxGpus(self, maxgpus):
        """Sets the default maximum number of gpus
        that new jobs are launched with.
        :type: float
        :param: value to set maxGpu to
        :rtype: show_pb2.ShowSetDefaultMaxGpuResponse
        :return: response is empty
        """
        response = self.stub.SetDefaultMaxGpus(show_pb2.ShowSetDefaultMaxGpusRequest(
            show=self.data, max_gpus=maxgpus),
            timeout=Cuebot.Timeout)
        return response

    def setDefaultMinGpus(self, mingpus):
        """Sets the default minimum number of gpus
        all new jobs are launched with.
        :type: float
        :param: value to set minGpus to
        :rtype: show_pb2.ShowSetDefaultMinGpusResponse
        :return: response is empty
        """
        response = self.stub.SetDefaultMinGpus(show_pb2.ShowSetDefaultMinGpusRequest(
            show=self.data, min_gpus=mingpus),
            timeout=Cuebot.Timeout)
        return response

    def findFilter(self, name):
        """Finds a filter by name.

        :type  name: string
        :param name: name of filter to find
        :rtype:  opencue.wrappers.filter.Filter
        :return: matching filter
        """
        response = self.stub.FindFilter(show_pb2.ShowFindFilterRequest(
            show=self.data, name=name), timeout=Cuebot.Timeout)
        return opencue.wrappers.filter.Filter(response.filter)

    def createFilter(self, name):
        """Creates a filter on the show.

        :type  name: str
        :param name: name of the filter to create
        :rtype:  opencue.wrappers.filter.Filter
        :return: the new filter object
        """
        response = self.stub.CreateFilter(show_pb2.ShowCreateFilterRequest(
            show=self.data, name=name), timeout=Cuebot.Timeout)
        return opencue.wrappers.filter.Filter(response.filter)

    def getGroups(self):
        """Gets the groups for the show.

        :rtype:  list<opencue.wrappers.group.Group>
        :return: list of groups for this show
        """
        response = self.stub.GetGroups(show_pb2.ShowGetGroupsRequest(
            show=self.data),
            timeout=Cuebot.Timeout)
        groupSeq = response.groups
        return [opencue.wrappers.group.Group(grp) for grp in groupSeq.groups]

    def getJobWhiteboard(self):
        """Gets the whiteboard for the show.

        :rtype:  job_pb2.NestedGroup
        :return: NestedGroup whiteboard for the show
        """
        response = self.stub.GetJobWhiteboard(show_pb2.ShowGetJobWhiteboardRequest(
            show=self.data),
            timeout=Cuebot.Timeout)
        return response.whiteboard

    def getRootGroup(self):
        """Gets the root group for the show.

        :rtype:  opencue.wrappers.group.Group
        :return: the root group
        """
        response = self.stub.GetRootGroup(show_pb2.ShowGetRootGroupRequest(
            show=self.data),
            timeout=Cuebot.Timeout)
        return opencue.wrappers.group.Group(response.group)

    def enableBooking(self, value):
        """Enables or disables booking on the show.

        :type  value: bool
        :param value: whether to enable booking
        :rtype:  show_pb2.ShowEnableBookingResponse
        :return: response is empty
        """
        response = self.stub.EnableBooking(show_pb2.ShowEnableBookingRequest(
            show=self.data,
            enabled=value),
            timeout=Cuebot.Timeout)
        return response

    def enableDispatching(self, value):
        """Enables or disables dispatching on the show.

        :type value: bool
        :param value: whether to enable booking
        :rtype:  show_pb2.ShowEnableDispatchingResponse
        :return: response is empty
        """
        response = self.stub.EnableDispatching(show_pb2.ShowEnableDispatchingRequest(
            show=self.data,
            enabled=value),
            timeout=Cuebot.Timeout)
        return response

    def id(self):
        """Returns the show id.

        :rtype:  str
        :return: id of the show
        """
        return self.data.id

    def name(self):
        """Returns the show name.

        :rtype:  str
        :return: name of the show
        """
        return self.data.name

    def pendingJobs(self):
        """Returns the total number of pending jobs on the show.

        :rtype:  int
        :return: total number of pending jobs
        """
        return self.data.show_stats.pending_jobs

    def pendingFrames(self):
        """Returns the total number of running frames currently in the queue.

        :rtype:  int
        :return: the total number of pending frames
        """
        return self.data.show_stats.pending_frames

    def runningFrames(self):
        """Returns the total number of running frames currently in the queue.

        :rtype:  int
        :return: the total number of running frames
        """
        return self.data.show_stats.running_frames

    def deadFrames(self):
        """Returns the total number of dead frames currently in the queue.

        :rtype:  int
        :return: the total number dead frames
        """
        return self.data.show_stats.dead_frames

    def reservedCores(self):
        """Returns the total number of reserved cores by all frames.

        :rtype:  float
        :return: the total number of reserved cores
        """
        return self.data.show_stats.reserved_cores

    def defaultMinProcs(self):
        """Returns the default minProcs that new jobs are set to.

        :rtype:  int
        :return: default minProcs value for new jobs
        """
        return self.data.default_min_procs

    def defaultMaxProcs(self):
        """Returns the default maxProcs that new jobs are set to.

        :rtype:  int
        :return: default maxProcs value for new jobs
        """
        return self.data.default_max_procs

    def totalJobsCreated(self):
        """A running counter of jobs launched.

        :rtype:  int
        :return: total number of jobs created
        """
        return self.data.show_stats.created_job_count

    def totalFramesCreated(self):
        """A running counter of frames launched.

        :rtype:  int
        :return: total number of frames created
        """
        return self.data.show_stats.created_frame_count



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module for classes related to dependencies."""

import enum

from opencue_proto import depend_pb2
from opencue.cuebot import Cuebot


class Depend(object):
    """This class contains the grpc implementation related to a Depend."""

    class DependType(enum.IntEnum):
        """Enum representing the type of dependency between subject and object."""
        JOB_ON_JOB = depend_pb2.JOB_ON_JOB
        JOB_ON_LAYER = depend_pb2.JOB_ON_LAYER
        JOB_ON_FRAME = depend_pb2.JOB_ON_FRAME
        LAYER_ON_JOB = depend_pb2.LAYER_ON_JOB
        LAYER_ON_LAYER = depend_pb2.LAYER_ON_LAYER
        LAYER_ON_FRAME = depend_pb2.LAYER_ON_FRAME
        FRAME_ON_JOB = depend_pb2.FRAME_ON_JOB
        FRAME_ON_LAYER = depend_pb2.FRAME_ON_LAYER
        FRAME_ON_FRAME = depend_pb2.FRAME_ON_FRAME
        FRAME_BY_FRAME = depend_pb2.FRAME_BY_FRAME
        PREVIOUS_FRAME = depend_pb2.PREVIOUS_FRAME
        LAYER_ON_SIM_FRAME = depend_pb2.LAYER_ON_SIM_FRAME

    class DependTarget(enum.IntEnum):
        """The type of target represented by this dependency."""
        INTERNAL = depend_pb2.INTERNAL
        EXTERNAL = depend_pb2.EXTERNAL
        ANY_TARGET = depend_pb2.ANY_TARGET

    def __init__(self, depend=None):
        self.data = depend
        self.stub = Cuebot.getStub('depend')

    def satisfy(self):
        """Satisfies the dependency.

        This sets any frames waiting on this dependency to the WAITING state.
        """
        self.stub.Satisfy(
            depend_pb2.DependSatisfyRequest(depend=self.data), timeout=Cuebot.Timeout)

    def unsatisfy(self):
        """Unsatisfies the dependency.

        This makes the dependency active again and sets matching frames to DEPEND.
        """
        self.stub.Unsatisfy(
            depend_pb2.DependUnsatisfyRequest(depend=self.data), timeout=Cuebot.Timeout)

    def id(self):
        """Returns the dependency's unique id.

        Dependencies are one of the only entities without a unique name so the unique ID is
        exposed to act as the name. This is mainly to make command line tools easier to use.

        :rtype:  str
        :return: the dependency's unique id"""
        return self.data.id

    def isInternal(self):
        """Returns whether the dependency is contained within a single job.

        :rtype:  bool
        :return: whether the dependency is contained within a single job
        """
        if self.data.depend_er_job == self.data.depend_on_job:
            return True
        return False

    def type(self):
        """Returns the type of dependency.

        :rtype: opencue_proto.depend_pb2.DependType
        :return: dependency type
        """
        return self.data.type

    def target(self):
        """Returns the target of the dependency, either internal or external.

        :rtype: opencue_proto.depend_pb2.DependTarget
        :return: dependency target type
        """
        return self.data.target

    def anyFrame(self):
        """Returns whether the depend is an any-frame depend.

        :rtype:  bool
        :return: whether the depend is an any-frame depend
        """
        return self.data.any_frame

    def isActive(self):
        """Returns whether the depend is active.

        :rtype:  bool
        :return: whether the depend is active
        """
        return self.data.active

    def dependErJob(self):
        """Returns the name of the job that is depending.

        :rtype:  str
        :return: name of the job that is depending"""
        return self.data.depend_er_job

    def dependErLayer(self):
        """Returns the name of the layer that is depending.

        :rtype:  str
        :return: name of the layer that is depending
        """
        return self.data.depend_er_layer

    def dependErFrame(self):
        """Returns the name of the frame that is depending.

        :rtype:  str
        :return: name of the frame that is depending
        """
        return self.data.depend_er_frame

    def dependOnJob(self):
        """Returns the name of the job to depend on.

        :rtype:  str
        :return: name of the job to depend on
        """
        return self.data.depend_on_job

    def dependOnLayer(self):
        """Returns the name of the layer to depend on.

        :rtype:  str
        :return: name of the layer to depend on
        """
        return self.data.depend_on_layer

    def dependOnFrame(self):
        """Returns the name of the frame to depend on.

        :rtype:  str
        :return: name of the frame to depend on
        """
        return self.data.depend_on_frame



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module for classes related to jobs."""

import enum
import getpass
import os
import platform
import time

from opencue_proto import comment_pb2
from opencue_proto import job_pb2
from opencue import Cuebot
import opencue.api
import opencue.search
import opencue.wrappers.comment
import opencue.wrappers.depend
import opencue.wrappers.frame
import opencue.wrappers.layer


class Job(object):
    """This class contains the ice implementation related to a job."""

    class JobState(enum.IntEnum):
        """Enum representing the state of a job."""
        PENDING = job_pb2.PENDING
        FINISHED = job_pb2.FINISHED
        STARTUP = job_pb2.STARTUP
        SHUTDOWN = job_pb2.SHUTDOWN
        POSTED = job_pb2.POSTED

    def __init__(self, job=None):
        self.data = job
        self.stub = Cuebot.getStub('job')
        self.__frameStateTotals = {}

    def kill(self, username=None, pid=None, host_kill=None, reason=None):
        """Kills the job."""
        username = username if username else getpass.getuser()
        pid = pid if pid else os.getpid()
        host_kill = host_kill if host_kill else platform.uname()[1]
        self.stub.Kill(job_pb2.JobKillRequest(job=self.data,
                                              username=username,
                                              pid=str(pid),
                                              host_kill=host_kill,
                                              reason=reason),
                       timeout=Cuebot.Timeout)

    def pause(self):
        """Pauses the job."""
        self.stub.Pause(job_pb2.JobPauseRequest(job=self.data), timeout=Cuebot.Timeout)

    def resume(self):
        """Resumes the job."""
        self.stub.Resume(job_pb2.JobResumeRequest(job=self.data), timeout=Cuebot.Timeout)

    def killFrames(self, username=None, pid=None, host_kill=None, reason=None, **request):
        """Kills all frames that match the FrameSearch.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        username = username if username else getpass.getuser()
        pid = pid if pid else os.getpid()
        host_kill = host_kill if host_kill else platform.uname()[1]
        criteria = opencue.search.FrameSearch.criteriaFromOptions(**request)
        self.stub.KillFrames(job_pb2.JobKillFramesRequest(job=self.data,
                                                          req=criteria,
                                                          username=username,
                                                          pid=str(pid),
                                                          host_kill=host_kill,
                                                          reason=reason),
                            timeout=Cuebot.Timeout)

    def eatFrames(self, **request):
        """Eats all frames that match the FrameSearch.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        criteria = opencue.search.FrameSearch.criteriaFromOptions(**request)
        return self.stub.EatFrames(job_pb2.JobEatFramesRequest(job=self.data, req=criteria),
                                   timeout=Cuebot.Timeout)

    def retryFrames(self, **request):
        """Retries all frames that match the FrameSearch.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        criteria = opencue.search.FrameSearch.criteriaFromOptions(**request)
        return self.stub.RetryFrames(job_pb2.JobRetryFramesRequest(job=self.data, req=criteria),
                                     timeout=Cuebot.Timeout)

    def markdoneFrames(self, **request):
        """Drops any dependency that requires any frame that matches the FrameSearch.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        criteria = opencue.search.FrameSearch.criteriaFromOptions(**request)
        return self.stub.MarkDoneFrames(
            job_pb2.JobMarkDoneFramesRequest(job=self.data, req=criteria),
            timeout=Cuebot.Timeout)

    def markAsWaiting(self, **request):
        """Changes the matching frames from the depend state to the waiting state.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        criteria = opencue.search.FrameSearch.criteriaFromOptions(**request)
        return self.stub.MarkAsWaiting(
            job_pb2.JobMarkAsWaitingRequest(job=self.data, req=criteria),
            timeout=Cuebot.Timeout)

    def setMinCores(self, minCores):
        """Sets the minimum number of cores the job needs.

        :type  minCores: int
        :param minCores: new minimum cores value
        """
        self.stub.SetMinCores(job_pb2.JobSetMinCoresRequest(job=self.data, val=minCores),
                              timeout=Cuebot.Timeout)

    def setMaxCores(self, maxCores):
        """Sets the maximum number of cores the job will use.

        :type  maxCores: int
        :param maxCores: new maximum cores value
        """
        self.stub.SetMaxCores(job_pb2.JobSetMaxCoresRequest(job=self.data, val=maxCores),
                              timeout=Cuebot.Timeout)

    def setMinGpus(self, minGpus):
        """Sets the minimum procs value
        :type  minGpus: int
        :param minGpus: New minimum cores value"""
        self.stub.SetMinGpus(job_pb2.JobSetMinGpusRequest(job=self.data, val=minGpus),
                             timeout=Cuebot.Timeout)

    def setMaxGpus(self, maxGpus):
        """Sets the maximum procs value
        :type  maxGpus: int
        :param maxGpus: New maximum cores value"""
        self.stub.SetMaxGpus(job_pb2.JobSetMaxGpusRequest(job=self.data, val=maxGpus),
                             timeout=Cuebot.Timeout)

    def setPriority(self, priority):
        """Sets the job priority.

        :type  priority: int
        :param priority: new job priority number
        """
        self.stub.SetPriority(job_pb2.JobSetPriorityRequest(job=self.data, val=priority),
                              timeout=Cuebot.Timeout)

    def setMaxRetries(self, maxRetries):
        """Sets the number of retries before a frame goes dead.

        :type  maxRetries: int
        :param maxRetries: new max retries
        """
        self.stub.SetMaxRetries(
            job_pb2.JobSetMaxRetriesRequest(job=self.data, max_retries=maxRetries),
            timeout=Cuebot.Timeout)

    def getLayers(self):
        """Returns the list of layers in the job.

        :rtype:  list<opencue.wrappers.layer.Layer>
        :return: list of layers in the job
        """
        response = self.stub.GetLayers(job_pb2.JobGetLayersRequest(job=self.data),
                                       timeout=Cuebot.Timeout)
        layerSeq = response.layers
        return [opencue.wrappers.layer.Layer(lyr) for lyr in layerSeq.layers]

    def getLayer(self, layerName):
        """ Returns the layer with the specified name
        :type:   layername: str
        :rtype:  opencue.wrappers.layer.Layer
        :return: specific layer in the job
        """
        return opencue.api.findLayer(self.name(), layerName)

    def getFrames(self, **options):
        """Returns the list of up to 1000 frames from within the job.

        For example::

            # Allowed: offset, limit, states+, layers+. frameset, changedate
            frames = job.getFrames(show=["edu","beo"],user="jwelborn")
            frames = job.getFrames(show="edu",shot="bs.012")

        :rtype:  list<opencue.wrappers.frame.Frame>
        :return: list of frames
        """
        criteria = opencue.search.FrameSearch.criteriaFromOptions(**options)
        response = self.stub.GetFrames(job_pb2.JobGetFramesRequest(job=self.data, req=criteria),
                                       timeout=Cuebot.Timeout)
        frameSeq = response.frames
        return [opencue.wrappers.frame.Frame(frm) for frm in frameSeq.frames]

    def getUpdatedFrames(self, lastCheck, layers=None):
        """Returns a list of state information for frames that have been recently updated.

        This includes any frames that have changed since the last update time as well as the
        current state of the job. If layer proxies are provided in the layers list, only frames
        from those layers will be returned.

        :type  lastCheck: int
        :param lastCheck: epoch when last updated
        :type  layers: list<job_pb2.Layer>
        :param layers: list of layers to check, empty list checks all
        :rtype:  job_pb2.JobGetUpdatedFramesResponse
        :return: job state and a list of updated frames
        """
        if layers is not None:
            layerSeq = job_pb2.LayerSeq()
            # pylint: disable=no-member
            layerSeq.layers.extend(layers)
            # pylint: enable=no-member
        else:
            layerSeq = None
        return self.stub.GetUpdatedFrames(
            job_pb2.JobGetUpdatedFramesRequest(job=self.data, last_check=lastCheck,
                                               layer_filter=layerSeq),
            timeout=Cuebot.Timeout)

    def setAutoEating(self, value):
        """Sets the job autoeat field.

        If set to true, any frames that would become dead will become eaten.

        :type  value: bool
        :param value: whether job should autoeat
        """
        self.stub.SetAutoEat(job_pb2.JobSetAutoEatRequest(job=self.data, value=value),
                             timeout=Cuebot.Timeout)

    def addRenderPartition(self, hostname, threads, max_cores, num_mem, max_gpus, max_gpu_memory):
        """Adds a render partition to the job.

        :type  hostname: str
        :param hostname: hostname of the partition
        :type  threads: int
        :param threads: number of threads of the partition
        :type  max_cores: int
        :param max_cores: max cores enabled for the partition
        :type  num_mem: int
        :param num_mem: amount of memory reserved for the partition
        :type  max_gpus: int
        :param max_gpus: max gpu cores enabled for the partition
        :type  max_gpu_memory: int
        :param max_gpu_memory: amount of gpu memory reserved for the partition
        """
        self.stub.AddRenderPartition(
            job_pb2.JobAddRenderPartRequest(job=self.data,
                                            host=hostname,
                                            threads=threads,
                                            max_cores=max_cores,
                                            max_memory=num_mem,
                                            max_gpus=max_gpus,
                                            max_gpu_memory=max_gpu_memory,
                                            username=os.getenv("USER", "unknown")))

    def getWhatDependsOnThis(self):
        """Returns a list of dependencies that depend directly on this job.

        :rtype:  list<opencue.wrappers.depend.Depend>
        :return: list of dependencies that depend directly on this job
        """
        response = self.stub.GetWhatDependsOnThis(
            job_pb2.JobGetWhatDependsOnThisRequest(job=self.data),
            timeout=Cuebot.Timeout)
        dependSeq = response.depends
        return [opencue.wrappers.depend.Depend(dep) for dep in dependSeq.depends]

    def getWhatThisDependsOn(self):
        """Returns a list of dependencies that this job depends on.

        :rtype:  list<opencue.wrappers.depend.Depend>
        :return: dependencies that this job depends on
        """
        response = self.stub.GetWhatThisDependsOn(
            job_pb2.JobGetWhatThisDependsOnRequest(job=self.data),
            timeout=Cuebot.Timeout)
        dependSeq = response.depends
        return [opencue.wrappers.depend.Depend(dep) for dep in dependSeq.depends]

    def getDepends(self):
        """Returns a list of all depends this job is involved with.

        :rtype:  list<opencue.wrappers.depend.Depend>
        :return: all depends involved with this job
        """
        response = self.stub.GetDepends(
            job_pb2.JobGetDependsRequest(job=self.data),
            timeout=Cuebot.Timeout)
        dependSeq = response.depends
        return [opencue.wrappers.depend.Depend(dep) for dep in dependSeq.depends]

    def dropDepends(self, target):
        """Drops the desired dependency target.

        :type  target: depend_pb2.DependTarget
        :param target: the desired dependency target to drop
        """
        return self.stub.DropDepends(job_pb2.JobDropDependsRequest(job=self.data, target=target),
                                     timeout=Cuebot.Timeout)

    def createDependencyOnJob(self, job):
        """Creates and returns a job-on-job dependency.

        :type  job: opencue.wrappers.job.Job
        :param job: the job you want this job to depend on
        :rtype:  opencue.wrappers.depend.Depend
        :return: the new dependency
        """
        response = self.stub.CreateDependencyOnJob(
            job_pb2.JobCreateDependencyOnJobRequest(job=self.data, on_job=job.data),
            timeout=Cuebot.Timeout)
        return opencue.wrappers.depend.Depend(response.depend)

    def createDependencyOnLayer(self, layer):
        """Create and return a job-on-layer dependency.

        :type  layer: opencue.wrappers.layer.Layer
        :param layer: the layer you want this job to depend on
        :rtype:  opencue.wrappers.Depend
        :return: the new dependency
        """
        response = self.stub.CreateDependencyOnLayer(
            job_pb2.JobCreateDependencyOnLayerRequest(job=self.data, layer=layer.data),
            timeout=Cuebot.Timeout)
        return opencue.wrappers.depend.Depend(response.depend)

    def createDependencyOnFrame(self, frame):
        """Creates and returns a job-on-frame dependency.

        :type  frame: opencue.wrappers.frame.Frame
        :param frame: the frame you want this job to depend on
        :rtype:  opencue.wrappers.depend.Depend
        :return: the new dependency
        """
        response = self.stub.CreateDependencyOnFrame(
            job_pb2.JobCreateDependencyOnFrameRequest(job=self.data, frame=frame.data),
            timeout=Cuebot.Timeout)
        return opencue.wrappers.depend.Depend(response.depend)

    # TODO(gregdenton) Is this needed? (Issue #71)
    # def unbookProcs(self, subs, number, kill=False):
    #     """Unbook procs off job from specified allocations
    #     :type  subs: list<Subscription>
    #     :param subs: The subscriptions to unbook from
    #     :type  number: int
    #     :param number: the number of virtual procs to unbook
    #     :type  kill: bool
    #     :param kill: wheather or not to kill the frames as well"""
    #     self.proxy.unbookProcs([a.proxy for a in subs],number,kill)

    def addComment(self, subject, message):
        """Appends a comment to the job's comment list.

        :type  subject: str
        :param subject: comment subject
        :type  message: str
        :param message: comment message body
        """
        comment = comment_pb2.Comment(
            user=os.getenv("USER", "unknown"),
            subject=subject,
            message=message or " ",
            timestamp=0)
        self.stub.AddComment(job_pb2.JobAddCommentRequest(job=self.data, new_comment=comment),
                             timeout=Cuebot.Timeout)

    def getComments(self):
        """Returns the job's comment list.

        :rtype: list<opencue.wrappers.comment.Comment>
        :return: the job's comment list
        """
        response = self.stub.GetComments(job_pb2.JobGetCommentsRequest(job=self.data),
                                         timeout=Cuebot.Timeout)
        commentSeq = response.comments
        return [opencue.wrappers.comment.Comment(cmt) for cmt in commentSeq.comments]

    def setGroup(self, group):
        """Sets the job to a new group.

        :type  group: opencue.wrappers.group.Group
        :param group: the group you want the job to be in
        """
        self.stub.SetGroup(job_pb2.JobSetGroupRequest(job=self.data, group_id=group.id()),
                           timeout=Cuebot.Timeout)

    def reorderFrames(self, frame_range, order):
        """Reorders the specified frame range on this job.

        :type  frame_range: string
        :param frame_range: The frame range to reorder
        :type  order: job_pb2.Order
        :param order: First, Last or Reverse
        """
        self.stub.ReorderFrames(
            job_pb2.JobReorderFramesRequest(job=self.data, range=frame_range, order=order),
            timeout=Cuebot.Timeout)

    def staggerFrames(self, frame_range, stagger):
        """Staggers the specified frame range on this job.

        :type  frame_range: string
        :param frame_range: the frame range to stagger
        :type  stagger: int
        :param stagger: the amount to stagger by
        """
        self.stub.StaggerFrames(
            job_pb2.JobStaggerFramesRequest(job=self.data, range=frame_range, stagger=stagger),
            timeout=Cuebot.Timeout)

    def addSubscriber(self, subscriber):
        """Adds email subscriber to status change for the job

        :type subscriber: string
        :param subscriber: email address to send update when the job finishes
        """
        self.stub.AddSubscriber(job_pb2.JobAddSubscriberRequest(job=self.data,
                                                                subscriber=subscriber))

    def facility(self):
        """Returns the facility that the job must run in.

        :rtype: str
        :return: the job facility
        """
        return self.data.facility

    def id(self):
        """Returns the id of the job.

        :rtype:  str
        :return: id of the job
        """
        return self.data.id

    def name(self):
        """Returns the name of the job.

        :rtype:  str
        :return: name of the job
        """
        return self.data.name

    def show(self):
        """Returns the show name of the job.

        :rtype:  str
        :return: show name of the job
        """
        return self.data.show

    def shot(self):
        """Returns the shot name of the job.

        :rtype:  str
        :return: shot name of the job
        """
        return self.data.shot

    def logDir(self):
        """Returns the path to the job log files.

        :rtype:  str
        :return: path of job log files
        """
        return self.data.log_dir

    def uid(self):
        """Returns the uid of the person who owns the job.

        :rtype:  Optional[int]
        :return: uid of job owner
        """
        return self.data.uid if self.data.HasField("uid") else None

    def user(self):
        """Returns the username of the person who owns the job.

        :rtype: str
        :return: username of job owner
        """
        return self.data.user

    def username(self):
        """Returns the username of the person who owns the job.

        :rtype: str
        :return: username of job owner"""
        return self.user()

    def state(self):
        """Returns the job state.

        :rtype:  job_pb2.JobState
        :return: job state
        """
        return self.data.state

    def priority(self):
        """Returns the job priority.

        :rtype:  int
        :return: job priority
        """
        return self.data.priority

    def minCores(self):
        """Returns the minimum number of cores the job needs.

        :rtype:  int
        :return: job's min cores
        """
        return self.data.min_cores

    def maxCores(self):
        """Returns the maximum number of cores the job will use.

        :rtype:  int
        :return: job's max cores
        """
        return self.data.max_cores

    def minGpus(self):
        """Returns the minimum number of gpus the job needs.
        :rtype:  int
        :return: job's min gpus
        """
        return self.data.min_gpus

    def maxGpus(self):
        """Returns the maximum number of gpus the job will use.
        :rtype:  int
        :return: job's max gpus
        """
        return self.data.max_gpus

    def os(self):
        """Returns the job's operating system.

        :rtype:  str
        :return: operating system name of the job
        """
        return self.data.os

    # pylint: disable=redefined-builtin
    def startTime(self, format=None):
        """Returns the job start time in the desired format.

        Examples:
            None                    => 1203634027
            "%m/%d %H:%M"           => 02/21 14:47
            "%a %b %d %H:%M:%S %Y"  => Thu Feb 21 14:47:07 2008

        See the format table at:
        https://docs.python.org/3/library/time.html

        :type  format: str
        :param format: desired time format
        :rtype:  int/str
        :return: job start time in epoch, or string version of that timestamp if format given"""
        if not format:
            return self.data.start_time
        return time.strftime(format, time.localtime(self.data.start_time))

    # pylint: disable=redefined-builtin
    def stopTime(self, format=None):
        """Returns the job stop time in the desired format.

        Examples:
            None                    => 1203634027
            "%m/%d %H:%M"           => 02/21 14:47
            "%a %b %d %H:%M:%S %Y"  => Thu Feb 21 14:47:07 2008

        See the format table at:
        https://docs.python.org/3/library/time.html

        :type  format: str
        :param format: desired time format
        :rtype:  int/str
        :return: job stop time in epoch, or string version of that timestamp if format given"""
        if not format:
            return self.data.stop_time
        return time.strftime(format, time.localtime(self.data.stop_time))

    def runTime(self):
        """Returns the number of seconds that the job has been (or was) running.

        :rtype:  int
        :return: job runtime in seconds
        """
        if self.data.stop_time == 0:
            return int(time.time() - self.data.start_time)
        return self.data.stop_time - self.data.start_time

    def coreSecondsRemaining(self):
        """Returns the estimated number of core seconds needed to finish all waiting frames.

        Does not take into account running frames.

        :rtype:  long
        :return: core seconds remaining
        """
        return self.data.job_stats.remaining_core_sec

    def age(self):
        """Returns the number of seconds since the job was launched.

        :rtype:  int
        :return: seconds since the job was launched
        """
        return int(time.time() - self.data.start_time)

    def isPaused(self):
        """Returns true if the job is paused.

        :rtype:  bool
        :return: paused or not paused"""
        return self.data.is_paused

    def isAutoEating(self):
        """Returns true if the job is eating all frames that become dead.

        :rtype:  bool
        :return: if the job eating all frames that become dead
        """
        return self.data.auto_eat

    def isCommented(self):
        """Returns true if the job has a comment.

        :rtype:  bool
        :return: if the job has a comment
        """
        return self.data.has_comment

    def setAutoEat(self, value):
        """Sets a new autoeat value for the job.

        Autoeat means job frames, when they would become dead, are eaten instead.

        :type  value: bool
        :param value: new state for autoeat
        """
        self.setAutoEating(value)
        self.data.auto_eat = value

    def coresReserved(self):
        """Returns the number of reserved cores the job has.

        :rtype:  float
        :return: number of reserved cores
        """
        return self.data.job_stats.reserved_cores

    def totalFrames(self):
        """Returns the total number of frames the job has.

        :rtype:  int
        :return: total number of frames
        """
        return self.data.job_stats.total_frames

    def totalLayers(self):
        """Returns the total number of layers the job has.

        :rtype:  int
        :return: total number of layers
        """
        return self.data.job_stats.total_layers

    def dependFrames(self):
        """Returns the total number of dependent frames the job has.

        :rtype:  int
        :return: total number of dependent frames
        """
        return self.data.job_stats.depend_frames

    def succeededFrames(self):
        """Returns the total number of succeeded frames the job has.

        :rtype:  int
        :return: total number of succeeded frames
        """
        return self.data.job_stats.succeeded_frames

    def runningFrames(self):
        """Returns the total number of running frames the job has.

        :rtype:  int
        :return: total number of running frames
        """
        return self.data.job_stats.running_frames

    def deadFrames(self):
        """Returns the total number of deads frames the job has.

        :rtype:  int
        :return: total number of dead frames
        """
        return self.data.job_stats.dead_frames

    def waitingFrames(self):
        """Returns the total number of waiting frames the job has.

        :rtype:  int
        :return: total number of waiting frames
        """
        return self.data.job_stats.waiting_frames

    def eatenFrames(self):
        """Returns the total number of eaten frames the job has.

        :rtype:  int
        :return: total number of eaten frames
        """
        return self.data.job_stats.eaten_frames

    def pendingFrames(self):
        """Returns the total number of pending (dependent and waiting) frames the job has.

        :rtype:  int
        :return: total number of pending (dependent and waiting) frames
        """
        return self.data.job_stats.pending_frames

    def frameStateTotals(self):
        """Returns a dictionary of frame states and the number of frames in each state.

        The states are available from job_pb2.FrameState.

        :rtype:  dict
        :return: total number of frames in each state
        """
        if not self.__frameStateTotals:
            self.__frameStateTotals.clear()
            for state in job_pb2.FrameState.keys():
                frameCount = getattr(self.data.job_stats, '{}_frames'.format(state.lower()), 0)
                self.__frameStateTotals[getattr(job_pb2, state)] = frameCount
        return self.__frameStateTotals

    def percentCompleted(self):
        """Returns the percent that the job's frames are completed.

        :rtype:  float
        :return: percentage of frame completion
        """
        try:
            return self.data.job_stats.succeeded_frames /\
                   float(self.data.job_stats.total_frames) * 100.0
        except ZeroDivisionError:
            return 0

    def group(self):
        """Returns the name of the group that the job is in.

        :rtype:  str
        :return: job group name
        """
        return self.data.group

    def avgFrameTime(self):
        """Returns the average completed frame time in seconds.

        :rtype:  int
        :return: average completed frame time in seconds
        """
        return self.data.job_stats.avg_frame_sec

    def averageCoreTime(self):
        """Returns the average frame core time.

        :rtype:  int
        :return: average frame core time for entire job
        """
        return self.data.job_stats.avg_core_sec

    def maxRss(self):
        """Returns the highest amount of memory that any frame in this job used.

        Value is within 5% of the actual highest frame.

        :rtype:  long
        :return: most memory used by any frame in kB"""
        return self.data.job_stats.max_rss

    def shutdownIfCompleted(self):
        """Shutdown the job if it is completed."""
        self.stub.ShutdownIfCompleted(job_pb2.JobShutdownIfCompletedRequest(job=self.data),
                                      timeout=Cuebot.Timeout)

    def lokiURL(self):
        """Returns url for loki server on the job

        :rtype: str
        :return: Return URL of loki server of the job
        """
        return self.data.loki_url

class NestedJob(Job):
    """This class contains information and actions related to a nested job."""
    def __init__(self, nestedJob=None):
        super(NestedJob, self).__init__(nestedJob)
        # job children are most likely empty but its possible to
        # populate this with NestedLayer objects.
        self.__children = []

    def children(self):
        """Returns all job children."""
        return self.__children

    def kill(self, username=None, pid=None, host_kill=None, reason=None):
        """Kills the job."""
        self.asJob().kill(username, pid, host_kill, reason)

    def pause(self):
        """Pauses the job."""
        self.asJob().pause()

    def resume(self):
        """Resumes the job."""
        self.asJob().resume()

    def killFrames(self, username=None, pid=None, host_kill=None, reason=None, **request):
        """Kills all frames that match the FrameSearch.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        self.asJob().killFrames(username, pid, host_kill, reason, **request)

    def eatFrames(self, **request):
        """Eats all frames that match the FrameSearch.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        self.asJob().eatFrames(**request)

    def retryFrames(self, **request):
        """Retries all frames that match the FrameSearch.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        self.asJob().retryFrames(**request)

    def markdoneFrames(self, **request):
        """Drops any dependency that requires any frame that matches the FrameSearch.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        self.asJob().markdoneFrames(**request)

    def markAsWaiting(self, **request):
        """Changes the matching frames from the depend state to the waiting state.

        :type  request: Dict
        :param request: FrameSearch parameters
        """
        self.asJob().markAsWaiting(**request)

    def setMinCores(self, minCores):
        """Sets the minimum cores value.

        :type  minCores: int
        :param minCores: new minimum cores value
        """
        self.asJob().setMinCores(minCores)

    def setMaxCores(self, maxCores):
        """Sets the maximum cores value.

        :type  maxCores: int
        :param maxCores: new maximum cores value
        """
        self.asJob().setMaxCores(maxCores)

    def setMinGpus(self, minGpus):
        """Sets the minimum gpus value
        :type  minGpus: int
        :param minGpus: New minimum gpus value"""
        self.asJob().setMinGpus(minGpus)

    def setMaxGpus(self, maxGpus):
        """Sets the maximum gpus value
        :type  maxGpus: int
        :param maxGpus: New maximum gpus value"""
        self.asJob().setMaxGpus(maxGpus)

    def setPriority(self, priority):
        """Sets the job priority.

        :type  priority: int
        :param priority: new priority number
        """
        self.asJob().setPriority(priority)

    def setMaxRetries(self, maxRetries):
        """Sets the number of retries before a frame goes dead.

        :type  maxRetries: int
        :param maxRetries: new max retries
        """
        self.asJob().setMaxRetries(maxRetries)

    def getLayers(self):
        """Returns the list of job layers.

        :rtype:  list<opencue.wrappers.layer.Layer>
        :return: list of layers in the job
        """
        return self.asJob().getLayers()

    def getFrames(self, **options):
        """Returns the list of up to 1000 frames from within the job.

        frames = job.getFrames(show=["edu","beo"],user="jwelborn")
        frames = job.getFrames(show="edu",shot="bs.012")
        Allowed: offset, limit, states+, layers+. frameset, changedate

        :rtype:  list<opencue.wrappers.frame.Frame>
        :return: list of matching frames"""
        return self.asJob().getFrames(**options)

    def getUpdatedFrames(self, lastCheck, layers=None):
        """Returns a list of state information for frames that have been recently updated.

        This includes any frames that have changed since the last update time as well as the
        current state of the job. If layer proxies are provided in the layers list, only frames
        from those layers will be returned.

        :type  lastCheck: int
        :param lastCheck: epoch when last updated
        :type  layers: list<job_pb2.Layer>
        :param layers: list of layers to check, empty list checks all
        :rtype:  job_pb2.JobGetUpdatedFramesResponse
        :return: job state and a list of updated frames
        """
        return self.asJob().getUpdatedFrames(lastCheck, layers)

    def setAutoEating(self, value):
        """If set to true, any frames that would become dead, will become eaten.

        :type  value: bool
        :param value: state of autoeat
        """
        self.asJob().setAutoEating(value)

    def getWhatDependsOnThis(self):
        """Returns a list of dependencies that depend directly on this job.

        :rtype:  list<opencue.wrappers.depend.Depend>
        :return: list of dependencies that depend directly on this job
        """
        return self.asJob().getWhatDependsOnThis()

    def getWhatThisDependsOn(self):
        """Returns a list of dependencies that this job depends on.

        :rtype:  list<opencue.wrappers.depend.Depend>
        :return: dependencies that this job depends on
        """
        return self.asJob().getWhatThisDependsOn()

    def getDepends(self):
        """Returns a list of all depends this job is involved with.

        :rtype:  list<opencue.wrappers.depend.Depend>
        :return: all depends involved with this job
        """
        return self.asJob().getDepends()

    def dropDepends(self, target):
        """Drops the desired dependency target.

            depend_pb2.DependTarget.AnyTarget
            depend_pb2.DependTarget.External
            depend_pb2.DependTarget.Internal

        :type  target: depend_pb2.DependTarget
        :param target: The desired dependency target to drop
        """
        return self.asJob().dropDepends(target)

    def createDependencyOnJob(self, job):
        """Creates and returns a job-on-job dependency.

        :type  job: opencue.wrappers.job.Job
        :param job: the job you want this job to depend on
        :rtype:  opencue.wrappers.depend.Depend
        :return: The new dependency
        """
        return self.asJob().createDependencyOnJob(job)

    def createDependencyOnLayer(self, layer):
        """Creates and returns a job-on-layer dependency.

        :type  layer: opencue.wrappers.layer.Layer
        :param layer: the layer you want this job to depend on
        :rtype:  opencue.wrappers.depend.Depend
        :return: the new dependency
        """
        return self.asJob().createDependencyOnLayer(layer)

    def createDependencyOnFrame(self, frame):
        """Creates and returns a job-on-frame dependency.

        :type  frame: opencue.wrappers.frame.Frame
        :param frame: the frame you want this job to depend on
        :rtype:  opencue.wrappers.depend.Depend
        :return: the new dependency
        """
        return self.asJob().createDependencyOnFrame(frame)

    def addComment(self, subject, message):
        """Appends a comment to the job's comment list.

        :type  subject: str
        :param subject: comment subject
        :type  message: str
        :param message: comment message body
        """
        self.asJob().addComment(subject, message)

    def getComments(self):
        """Returns the job's comment list."""
        return self.asJob().getComments()

    def setGroup(self, group):
        """Sets the job to a new group.

        :type  group: opencue.wrappers.group.Group
        :param group: the group you want the job to be in.
        """
        self.asJob().setGroup(group)

    def reorderFrames(self, frame_range, order):
        """Reorders the specified frame range on this job.

        :type  frame_range: string
        :param frame_range: The frame range to reorder
        :type  order: job_pb2.Order
        :param order: First, Last or Reverse
        """
        self.asJob().reorderFrames(frame_range, order)

    def staggerFrames(self, frame_range, stagger):
        """Staggers the specified frame range on this job.

        :type  frame_range: string
        :param frame_range: the frame range to stagger
        :type  stagger: int
        :param stagger: the amount to stagger by
        """
        self.asJob().staggerFrames(frame_range, stagger)

    def asJob(self):
        """Returns a Job object from this NestedJob.

        :rtype: opencue.wrappers.job.Job
        :return: Job version of this NestedJob
        """
        return Job(job_pb2.Job(
            id=self.data.id,
            state=self.data.state,
            name=self.data.name,
            shot=self.data.shot,
            show=self.data.show,
            user=self.data.user,
            group=self.data.group,
            facility=self.data.facility,
            os=self.data.os,
            uid=self.data.uid if self.data.HasField("uid") else None,
            priority=self.data.priority,
            min_cores=self.data.min_cores,
            max_cores=self.data.max_cores,
            log_dir=self.data.log_dir,
            is_paused=self.data.is_paused,
            has_comment=self.data.has_comment,
            auto_eat=self.data.auto_eat,
            start_time=self.data.start_time,
            stop_time=self.data.stop_time,
            job_stats=self.data.stats
        ))



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Tests for `opencue.util`."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
import unittest
import uuid

import grpc
import mock

from opencue_proto import job_pb2
import opencue.exception
from opencue.wrappers.job import Job


@opencue.util.grpcExceptionParser
def testRaise(exc):
    uuid.uuid4()
    raise exc


class ProxyTests(unittest.TestCase):
    """proxy converts different types of entities to usable Ice proxies"""

    def testNotFoundExceptionParser(self):
        response = grpc.RpcError()
        response.code = lambda: grpc.StatusCode.NOT_FOUND
        response.details = lambda: "testing not found"
        self.assertRaises(opencue.exception.EntityNotFoundException,
                          lambda: testRaise(response))

    def testExistsExceptionParser(self):
        response = grpc.RpcError()
        response.code = lambda: grpc.StatusCode.ALREADY_EXISTS
        response.details = lambda: "already exists"
        self.assertRaises(opencue.exception.EntityAlreadyExistsException,
                          lambda: testRaise(response))

    def testDeadlineExceptionParser(self):
        response = grpc.RpcError()
        response.code = lambda: grpc.StatusCode.DEADLINE_EXCEEDED
        response.details = lambda: "deadline exceeded"
        self.assertRaises(opencue.exception.DeadlineExceededException,
                          lambda: testRaise(response))

    def testInternalExceptionParser(self):
        response = grpc.RpcError()
        response.code = lambda: grpc.StatusCode.INTERNAL
        response.details = lambda: "Internal Error"
        self.assertRaises(opencue.exception.CueInternalErrorException,
                          lambda: testRaise(response))

    @mock.patch('uuid.uuid4')
    def testConnectionExceptionParser(self, mockUuid):
        with mock.patch('opencue.exception.getRetryCount', return_value=3):
            response = grpc.RpcError()
            response.code = lambda: grpc.StatusCode.UNAVAILABLE
            response.details = lambda: "Connection Error"
            self.assertRaises(opencue.exception.ConnectionException,
                              lambda: testRaise(response))
        self.assertEqual(mockUuid.call_count, 4)

    @mock.patch('uuid.uuid4')
    def testConnectionExceptionZeroRetries(self, mockUuid):
        with mock.patch('opencue.exception.getRetryCount', return_value=0):
            response = grpc.RpcError()
            response.code = lambda: grpc.StatusCode.UNAVAILABLE
            response.details = lambda: "Connection Error"
            self.assertRaises(opencue.exception.ConnectionException,
                              lambda: testRaise(response))
        self.assertEqual(mockUuid.call_count, 1)

    def testUnknownExceptionParser(self):
        response = grpc.RpcError()
        response.code = lambda: "unknown"
        response.details = lambda: "unknown"
        self.assertRaises(opencue.exception.CueException,
                          lambda: testRaise(response))

    @mock.patch('opencue.cuebot.Cuebot.getStub')
    def testProxyUniqueId(self, getStubMock):
        """convert a string and class name to proxy"""
        objectId = 'A0000000-0000-0000-0000-000000000000'
        stubMock = mock.Mock()
        stubMock.GetGroup.return_value = job_pb2.GroupGetGroupResponse(
            group=job_pb2.Group(id=objectId))
        getStubMock.return_value = stubMock

        proxy = opencue.proxy(objectId, 'Group')

        stubMock.GetGroup.assert_called_with(job_pb2.GroupGetGroupRequest(id=objectId))
        self.assertEqual(objectId, proxy.group.id)

    @mock.patch('opencue.cuebot.Cuebot.getStub')
    def testProxyUniqueIdArray(self, getStubMock):
        """convert a list of strings and a class name to a proxy"""
        ids = ['A0000000-0000-0000-0000-000000000000', 'B0000000-0000-0000-0000-000000000000']
        stubMock = mock.Mock()
        stubMock.GetGroup.side_effect = lambda request : job_pb2.GroupGetGroupResponse(
            group=job_pb2.Group(id=request.id))
        getStubMock.return_value = stubMock

        proxyList = opencue.proxy(ids, 'Group')

        stubMock.GetGroup.assert_has_calls([
            mock.call(job_pb2.GroupGetGroupRequest(id=ids[0])),
            mock.call(job_pb2.GroupGetGroupRequest(id=ids[1])),
        ])
        self.assertEqual(ids, [proxy.group.id for proxy in proxyList])

    @mock.patch('opencue.cuebot.Cuebot.getStub')
    def testProxyObjectArray(self, getStubMock):
        """convert a list of proto object and a class name to a proxy"""
        ids = ['A0000000-0000-0000-0000-000000000000', 'B0000000-0000-0000-0000-000000000000']
        protos = [job_pb2.Group(id=id) for id in ids]
        stubMock = mock.Mock()
        stubMock.GetGroup.side_effect = lambda request: job_pb2.GroupGetGroupResponse(
            group=job_pb2.Group(id=request.id))
        getStubMock.return_value = stubMock

        proxyList = opencue.proxy(protos, 'Group')

        stubMock.GetGroup.assert_has_calls([
            mock.call(job_pb2.GroupGetGroupRequest(id=ids[0])),
            mock.call(job_pb2.GroupGetGroupRequest(id=ids[1])),
        ])
        self.assertEqual(ids, [proxy.group.id for proxy in proxyList])


class IdTests(unittest.TestCase):
    """id() takes an entity and returns the unique id"""

    @mock.patch('opencue.cuebot.Cuebot.getStub')
    def testIdOnEntity(self, getStubMock):
        del getStubMock
        arbitraryId = 'foo'
        job = Job(job_pb2.Job(id=arbitraryId))

        self.assertEqual(arbitraryId, opencue.id(job))

    @mock.patch('opencue.cuebot.Cuebot.getStub')
    def testIdOnEntityList(self, getStubMock):
        del getStubMock
        arbitraryIds = ['foo', 'bar']
        jobs = [Job(job_pb2.Job(id=arbitraryIds[0])), Job(job_pb2.Job(id=arbitraryIds[1]))]

        ids = opencue.id(jobs)

        self.assertEqual(len(jobs), len(ids))
        for i in range(0, len(jobs)):
            self.assertEqual(arbitraryIds[i], ids[i])


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module for classes related to limits."""

from opencue_proto import limit_pb2
from opencue import Cuebot


class Limit(object):
    """This class contains the grpc implementation related to a Limit."""

    def __init__(self, limit=None):
        self.data = limit
        self.stub = Cuebot.getStub('limit')

    def create(self):
        """Creates a new Limit from the current Limit object.

        :rtype:  opencue.wrappers.limit.Limit
        :return: the newly created Limit
        """
        return Limit(self.stub.Create(
            limit_pb2.LimitCreateRequest(name=self.name(), max_value=self.maxValue()),
            timeout=Cuebot.Timeout))

    def delete(self):
        """Deletes the limit."""
        self.stub.Delete(limit_pb2.LimitDeleteRequest(name=self.name()), timeout=Cuebot.Timeout)

    def find(self, name):
        """Finds an existing limit by its name.

        :type  name: str
        :param name: name of limit to find
        :rtype:  opencue.wrappers.limit.Limit
        :return: the limit found by name
        """
        return Limit(
            self.stub.Find(limit_pb2.LimitFindRequest(name=name), timeout=Cuebot.Timeout).limit)

    def get(self, limit_id):
        """Returns an existing limit by its id.

        :type  limit_id: str
        :param limit_id: id of limit to find
        :rtype:  opencue.wrappers.limit.Limit
        :return: the limit found by id.
        """
        return Limit(
            self.stub.Get(limit_pb2.LimitGetRequest(id=limit_id), timeout=Cuebot.Timeout).limit)

    def rename(self, newName):
        """Renames the limit.

        :type  newName: str
        :param newName: new limit name
        """
        self.stub.Rename(limit_pb2.LimitRenameRequest(old_name=self.name(), new_name=newName),
                         timeout=Cuebot.Timeout)
        self._update()

    def setMaxValue(self, maxValue):
        """Sets the maximum value of an existing limit.

        :type  maxValue: int
        :param maxValue: new limit maximum
        """
        self.stub.SetMaxValue(
            limit_pb2.LimitSetMaxValueRequest(name=self.name(), max_value=maxValue),
            timeout=Cuebot.Timeout)
        self._update()

    def _update(self):
        """Updates the current data object from the database."""
        self.data = self.stub.Get(limit_pb2.LimitGetRequest(id=self.id()), timeout=Cuebot.Timeout)

    def id(self):
        """Returns the limit id.

        :rtype:  str
        :return: the limit id
        """
        return self.data.id

    def name(self):
        """Returns the limit name.

        :rtype:  str
        :return: the limit name
        """
        if hasattr(self.data, 'name'):
            return self.data.name
        return ""

    def maxValue(self):
        """Returns the limit maximum.

        :rtype: int
        :return: the limit maximum
        """
        if hasattr(self.data, 'max_value'):
            return self.data.max_value
        return -1

    def currentRunning(self):
        """Returns the current amount of the limit in use.

        :rtype: int
        :return: current limit usage
        """
        if hasattr(self.data, 'current_running'):
            return self.data.current_running
        return -1



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Tests for `opencue.wrappers.limit`."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import unittest

import mock

from opencue_proto import limit_pb2
import opencue.wrappers.limit


TEST_LIMIT_ID = 'lll-llll-lll'
TEST_LIMIT_NAME = 'imalimit'
TEST_LIMIT_MAX_VALUE = 42


@mock.patch('opencue.cuebot.Cuebot.getStub')
class LimitTests(unittest.TestCase):

    def testCreate(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.Create.return_value = limit_pb2.LimitCreateResponse()
        getStubMock.return_value = stubMock

        limit = opencue.wrappers.limit.Limit(
            limit_pb2.Limit(name=TEST_LIMIT_NAME, max_value=TEST_LIMIT_MAX_VALUE))
        limit.create()

        stubMock.Create.assert_called_with(
            limit_pb2.LimitCreateRequest(name=TEST_LIMIT_NAME, max_value=TEST_LIMIT_MAX_VALUE),
            timeout=mock.ANY)

    def testDelete(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.Delete.return_value = limit_pb2.LimitDeleteResponse()
        getStubMock.return_value = stubMock

        limit = opencue.wrappers.limit.Limit(
            limit_pb2.Limit(name=TEST_LIMIT_NAME, max_value=TEST_LIMIT_MAX_VALUE))
        limit.delete()

        stubMock.Delete.assert_called_with(
            limit_pb2.LimitDeleteRequest(name=TEST_LIMIT_NAME), timeout=mock.ANY)

    def testFind(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.Find.return_value = limit_pb2.LimitFindResponse(
          limit=limit_pb2.Limit(name=TEST_LIMIT_NAME, max_value=TEST_LIMIT_MAX_VALUE))
        getStubMock.return_value = stubMock

        limit = opencue.wrappers.limit.Limit()
        responseLimit = limit.find(TEST_LIMIT_NAME)

        stubMock.Find.assert_called_with(
            limit_pb2.LimitFindRequest(name=TEST_LIMIT_NAME), timeout=mock.ANY)
        self.assertEqual(responseLimit.name(), TEST_LIMIT_NAME)
        self.assertEqual(responseLimit.maxValue(), TEST_LIMIT_MAX_VALUE)

    def testGet(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.Get.return_value = limit_pb2.LimitGetResponse(
            limit=limit_pb2.Limit(name=TEST_LIMIT_NAME, max_value=TEST_LIMIT_MAX_VALUE))
        getStubMock.return_value = stubMock

        limit = opencue.wrappers.limit.Limit()
        responseLimit = limit.get(TEST_LIMIT_ID)

        stubMock.Get.assert_called_with(
            limit_pb2.LimitGetRequest(id=TEST_LIMIT_ID), timeout=mock.ANY)
        self.assertEqual(responseLimit.name(), TEST_LIMIT_NAME)
        self.assertEqual(responseLimit.maxValue(), TEST_LIMIT_MAX_VALUE)

    def testRename(self, getStubMock):
        test_new_name = 'new_name'
        stubMock = mock.Mock()
        stubMock.Rename.return_value = limit_pb2.LimitRenameResponse()
        getStubMock.return_value = stubMock

        limit = opencue.wrappers.limit.Limit(
            limit_pb2.Limit(name=TEST_LIMIT_NAME, max_value=TEST_LIMIT_MAX_VALUE))
        limit.rename(test_new_name)

        stubMock.Rename.assert_called_with(
            limit_pb2.LimitRenameRequest(old_name=TEST_LIMIT_NAME, new_name=test_new_name),
            timeout=mock.ANY)

    def testSetMaxValue(self, getStubMock):
        max_value = 16
        stubMock = mock.Mock()
        stubMock.SetMaxValue.return_value = limit_pb2.LimitSetMaxValueResponse()
        getStubMock.return_value = stubMock

        limit = opencue.wrappers.limit.Limit(
            limit_pb2.Limit(name=TEST_LIMIT_NAME, max_value=TEST_LIMIT_MAX_VALUE))
        limit.setMaxValue(max_value)

        stubMock.SetMaxValue.assert_called_with(
            limit_pb2.LimitSetMaxValueRequest(name=TEST_LIMIT_NAME, max_value=max_value),
            timeout=mock.ANY)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Tests for `opencue.wrappers.deed`."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import unittest

import mock

from opencue_proto import host_pb2
import opencue.wrappers.deed


TEST_DEED_ID = 'ddd-dd-dddd'
TEST_DEED_OWNER = 'testDeedOwner'
TEST_HOST_ID = 'hhh-hh-hhhh'


@mock.patch('opencue.cuebot.Cuebot.getStub')
class DeedTests(unittest.TestCase):

    def testDelete(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.Delete.return_value = host_pb2.DeedDeleteResponse()
        getStubMock.return_value = stubMock

        deed = opencue.wrappers.deed.Deed(host_pb2.Deed(id=TEST_DEED_ID))
        deed.delete()

        stubMock.Delete.assert_called_with(
            host_pb2.DeedDeleteRequest(deed=deed.data), timeout=mock.ANY)

    def testGetHost(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.GetHost.return_value = host_pb2.DeedGetHostResponse(
            host=host_pb2.Host(id=TEST_HOST_ID))
        getStubMock.return_value = stubMock

        deed = opencue.wrappers.deed.Deed(host_pb2.Deed(id=TEST_DEED_ID))
        host = deed.getHost()

        stubMock.GetHost.assert_called_with(
            host_pb2.DeedGetHostRequest(deed=deed.data), timeout=mock.ANY)
        self.assertEqual(host.id(), TEST_HOST_ID)

    def testGetOwner(self, getStubMock):
        stubMock = mock.Mock()
        stubMock.GetOwner.return_value = host_pb2.DeedGetOwnerResponse(
            owner=host_pb2.Owner(name=TEST_DEED_OWNER))
        getStubMock.return_value = stubMock

        deed = opencue.wrappers.deed.Deed(host_pb2.Deed(id=TEST_DEED_ID))
        owner = deed.getOwner()

        stubMock.GetOwner.assert_called_with(
            host_pb2.DeedGetOwnerRequest(deed=deed.data), timeout=mock.ANY)
        self.assertEqual(owner.name(), TEST_DEED_OWNER)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Tests for `FileSequence`."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import str
import unittest

from FileSequence import FrameRange
from FileSequence import FrameSet
from FileSequence import FileSequence


class FrameRangeTests(unittest.TestCase):

    def testSingleFrame(self):
        frame = 4927

        result = FrameRange(str(frame))

        self.assertEqual([frame], result.getAll())

    def testNegativeSingleFrame(self):
        frame = -4982

        result = FrameRange(str(frame))

        self.assertEqual([frame], result.getAll())

    def testFrameRange(self):
        result = FrameRange('1-7')

        self.assertEqual([1, 2, 3, 4, 5, 6, 7], result.getAll())

    def testNegativeFrameRange(self):
        result = FrameRange('-20--13')

        self.assertEqual([-20, -19, -18, -17, -16, -15, -14, -13], result.getAll())

    def testNegativeToPositiveFrameRange(self):
        result = FrameRange('-5-3')

        self.assertEqual([-5, -4, -3, -2, -1, 0, 1, 2, 3], result.getAll())

    def testReverseFrameRange(self):
        result = FrameRange('6-2')

        self.assertEqual([6, 5, 4, 3, 2], result.getAll())

    def testReverseNegativeFrameRange(self):
        result = FrameRange('-2--6')

        self.assertEqual([-2, -3, -4, -5, -6], result.getAll())

    def testStep(self):
        result = FrameRange('1-8x2')

        self.assertEqual([1, 3, 5, 7], result.getAll())

    def testNegativeStep(self):
        result = FrameRange('8-1x-2')

        self.assertEqual([8, 6, 4, 2], result.getAll())

    def testNegativeStepInvalidRange(self):
        with self.assertRaises(ValueError):
            FrameRange('1-8x-2')

    def testInvertedStep(self):
        result = FrameRange('1-8y2')

        self.assertEqual([2, 4, 6, 8], result.getAll())

    def testNegativeInvertedStep(self):
        result = FrameRange('8-1y-2')

        self.assertEqual([7, 5, 3, 1], result.getAll())

    def testInterleave(self):
        result = FrameRange('1-10:5')

        self.assertEqual([1, 6, 2, 4, 8, 10, 3, 5, 7, 9], result.getAll())

    def testNegativeInterleave(self):
        result = FrameRange('10-1:-5')

        self.assertEqual([10, 5, 9, 7, 3, 1, 8, 6, 4, 2], result.getAll())

    def testNonNumericalInput(self):
        with self.assertRaises(ValueError):
            FrameRange('a')

        with self.assertRaises(ValueError):
            FrameRange('a-b')

        with self.assertRaises(ValueError):
            FrameRange('1-5xc')

        with self.assertRaises(ValueError):
            FrameRange('1-5:c')

    def testInvalidRange(self):
        with self.assertRaises(ValueError):
            FrameRange('1-10-20')

        with self.assertRaises(ValueError):
            FrameRange('1x10-20')

        with self.assertRaises(ValueError):
            FrameRange('1:10-20')

    def testSize(self):
        result = FrameRange('1-7')

        self.assertEqual(7, result.size())

    def testGet(self):
        result = FrameRange('1-7')

        self.assertEqual(5, result.get(4))

    def testIndex(self):
        result = FrameRange('1-7')

        self.assertEqual(5, result.index(6))
        self.assertEqual(-1, result.index(22))

    def testNormalize(self):
        simpleRange = FrameRange('3-5')
        simpleRange.normalize()

        self.assertEqual([3, 4, 5], simpleRange.getAll())

        reverseOrder = FrameRange('3-1x-1')
        reverseOrder.normalize()

        self.assertEqual([1, 2, 3], reverseOrder.getAll())


class FrameSetTests(unittest.TestCase):

    def testMultipleSegments(self):
        result = FrameSet('57,1-3,4-2,12-15x2,76-70x-3,5-12y3,1-7:5')

        self.assertEqual(
            [57, 1, 2, 3, 4, 3, 2, 12, 14, 76, 73, 70, 6, 7, 9, 10, 12, 1, 6, 2, 4, 3, 5, 7],
            result.getAll())

    def testSize(self):
        result = FrameSet('1-7')

        self.assertEqual(7, result.size())

    def testGet(self):
        result = FrameSet('1-7')

        self.assertEqual(5, result.get(4))

    def testIndex(self):
        result = FrameSet('1-7')

        self.assertEqual(5, result.index(6))
        self.assertEqual(-1, result.index(22))

    def testNormalize(self):
        reverseOrder = FrameSet('5,3-1x-1')
        reverseOrder.normalize()

        self.assertEqual([1, 2, 3, 5], reverseOrder.getAll())

        duplicates = FrameSet('1-2,2-3')
        duplicates.normalize()

        self.assertEqual([1, 2, 3], duplicates.getAll())


class FileSequenceTests(unittest.TestCase):
    def __testFileSequence(self, filespec, **kwargs):
        fs = FileSequence(filespec)

        tests = {'prefix': fs.getPrefix(),
                 'frameSet': fs.frameSet,
                 'suffix': fs.getSuffix(),
                 'padSize': fs.getPadSize(),
                 'dirname': fs.getDirname(),
                 'basename': fs.getBasename()
                 }

        for arg, member in tests.items():
            if arg in kwargs:
                if isinstance(member, FrameSet):
                    self.assertEqual(member.getAll(), kwargs[arg].getAll(),
                                     "Comparing '%s', got '%s', expected '%s'" % (arg, str(member),
                                                                                  str(kwargs[arg])))
                else:
                    self.assertEqual(member, kwargs[arg],
                                     "Comparing '%s', got '%s', expected '%s'" % (arg, str(member),
                                                                                  str(kwargs[arg])))

    def testVariousFileSequences(self):
        """Test various file sequences are correctly parsed."""
        self.__testFileSequence('foo.1-1####.bar', prefix='foo.', frameSet=FrameSet('1-1'),
                                suffix='.bar', padSize=4)
        self.__testFileSequence('foo.####.bar', prefix='foo.', frameSet=None, suffix='.bar',
                                padSize=4)
        # Not sure why this becomes padSize of 10
        # self.__testFileSequence('foo.1-15x2#@#@.bar', prefix='foo.', frameSet=FrameSet('1-15x2'),
        # suffix='.bar',
        # padSize=10)
        self.__testFileSequence('foo.1-15x2.bar', prefix='foo.', frameSet=FrameSet('1-15x2'),
                                suffix='.bar', padSize=1)
        self.__testFileSequence('someImage.1,3,5####.rla', prefix='someImage.',
                                frameSet=FrameSet('1,3,5'), suffix='.rla', padSize=4)
        self.__testFileSequence('foo.####.exr.tx', prefix='foo.', frameSet=None, suffix='.exr.tx',
                                padSize=4)
        self.__testFileSequence('foo.1-10#.bar.1-9####.bar', prefix='foo.1-10#.bar.',
                                frameSet=FrameSet('1-9'), suffix='.bar', padSize=4)
        self.__testFileSequence('foo.1-9.bar', prefix='foo.', frameSet=FrameSet('1-9'),
                                suffix='.bar', padSize=1)
        self.__testFileSequence('foo.1-10.bar', prefix='foo.', frameSet=FrameSet('1-10'),
                                suffix='.bar', padSize=1)
        self.__testFileSequence('foo.9.bar', prefix='foo.', frameSet=FrameSet('9-9'), suffix='.bar',
                                padSize=1)

        self.__testFileSequence('foo.1-10#.bar', prefix='foo.', dirname='', basename='foo')
        self.__testFileSequence('/foo.1-10#.bar', prefix='/foo.', dirname='/', basename='foo')
        self.__testFileSequence('baz/foo.1-10#.bar', prefix='baz/foo.', dirname='baz/',
                                basename='foo')
        self.__testFileSequence('/baz/foo.1-10#.bar', prefix='/baz/foo.', dirname='/baz/',
                                basename='foo')
        self.__testFileSequence('/bar/baz/foo.1-10#.bar', prefix='/bar/baz/foo.',
                                dirname='/bar/baz/', basename='foo')

        self.__testFileSequence('foo.-15-15####.bar', prefix='foo.', frameSet=FrameSet('-15-15'),
                                suffix='.bar', padSize=4)
        self.__testFileSequence('foo.-15--1####.bar', prefix='foo.', frameSet=FrameSet('-15--1'),
                                suffix='.bar', padSize=4)

    def testPadSizeWithoutPadTokens(self):
        """Test the pad size is correctly guessed when no padding tokens are given."""
        self.__testFileSequence('foo.0009.bar', padSize=4)
        self.__testFileSequence('foo.1-9x0002.bar', padSize=1)
        # This test contradicts another test for negative steps
        # self.__testFileSequence('foo.9-1x-0002.bar',    padSize=1)
        self.__testFileSequence('foo.9-09x0002.bar', padSize=1)
        self.__testFileSequence('foo.9,10.bar', padSize=1)
        self.__testFileSequence('foo.009,10.bar', padSize=3)
        self.__testFileSequence('foo.-011.bar', padSize=4)

        # sequence padded to 4 but frame count goes above 9999
        self.__testFileSequence('foo.0001-10000.bar', padSize=4)

    def testInvalidSequences(self):
        """Test invalid file sequences throw expected exception."""
        self.assertRaises(ValueError, FileSequence, 'asdasdasda')
        self.assertRaises(ValueError, FileSequence, 'foo.fred#.bar')
        self.assertRaises(ValueError, FileSequence, 'foo..bar')
        self.assertRaises(ValueError, FileSequence, 'foo.-,x#.bar')
        self.assertRaises(ValueError, FileSequence, 'foo.x2.bar')
        self.assertRaises(ValueError, FileSequence, 'foo.-20---10.bar')
        # order reversed
        self.assertRaises(ValueError, FileSequence, 'foo.10-1.bar')
        self.assertRaises(ValueError, FileSequence, 'foo.-10--20.bar')
        # require a prefix
        self.assertRaises(ValueError, FileSequence, '.1')
        self.assertRaises(ValueError, FileSequence, '0.1')

    def __testStringify(self, filespec, index, expected):
        fs = FileSequence(filespec)
        self.assertEqual(expected, fs[index])

    def testStringify(self):
        self.__testStringify('foo.011.bar', 0, 'foo.011.bar')
        self.__testStringify('foo.-011.bar', 0, 'foo.-011.bar')

    def __testFrameList(self, filespec, frame, expected):
        fs = FileSequence(filespec)
        self.assertEqual(expected, fs(frame))

    def testFrameList(self):
        self.__testFrameList('foo.1-10.bar', 4, 'foo.4.bar')
        self.__testFrameList('foo.1-10####.bar', 4, 'foo.0004.bar')
        self.__testFrameList('foo.####.bar', 4, 'foo.0004.bar')


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""OpenCue configuration."""

import logging
import os
import platform

import yaml


logger = logging.getLogger("opencue")


# Config file from which default settings are loaded. This file is distributed with the
# opencue Python library.
__DEFAULT_CONFIG_FILE = os.path.join(os.path.dirname(__file__), 'default.yaml')

# Environment variables which can be used to define a custom config file. Any settings
# defined in this file will be used instead of the defaults.
__CONFIG_FILE_ENV_VARS = [
    # OPENCUE_CONFIG_FILE is the preferred setting to use.
    'OPENCUE_CONFIG_FILE',
    # OPENCUE_CONF is deprecated, but kept for now for backwards compatibility.
    'OPENCUE_CONF',
]


def config_base_directory():
    """Returns the OpenCue config base directory.

    This platform-dependent directory, stored within your user profile, is used by
    OpenCue components as the default location for various configuration files. Typically
    if you store your config files in this location, there is no need to set environment
    variables to indicate where your config files are located -- OpenCue should recognize
    them automatically.

    NOTE: This work is ongoing. Over time more OpenCue components will start using this
    base directory. See https://github.com/AcademySoftwareFoundation/OpenCue/issues/785.

    :rtype: str
    :return: config file base directory
    """
    if platform.system() == 'Windows':
        return os.path.join(os.path.expandvars('%APPDATA%'), 'opencue')
    return os.path.join(os.path.expanduser('~'), '.config', 'opencue')


def load_config_from_file():
    """Loads configuration settings from config file on the local system.

    Default settings are read from default.yaml which is distributed with the opencue library.
    User-provided config is then read from disk, in order of preference:
    - Path defined by the OPENCUE_CONFIG_FILE environment variable.
    - Path defined by the OPENCUE_CONF environment variable.
    - Path within the config base directory (i.e. ~/.config/opencue/opencue.yaml)

    :rtype: dict
    :return: config settings
    """
    with open(__DEFAULT_CONFIG_FILE, encoding="utf-8") as file_object:
        config = yaml.load(file_object, Loader=yaml.SafeLoader)

    user_config_file = None

    for config_file_env_var in __CONFIG_FILE_ENV_VARS:
        logger.debug('Checking for opencue config file path in %s', config_file_env_var)
        config_file_from_env = os.environ.get(config_file_env_var)
        if config_file_from_env and os.path.exists(config_file_from_env):
            user_config_file = config_file_from_env
            break

    if not user_config_file:
        config_from_user_profile = os.path.join(config_base_directory(), 'opencue.yaml')
        logger.debug('Checking for opencue config at %s', config_from_user_profile)
        if os.path.exists(config_from_user_profile):
            user_config_file = config_from_user_profile

    if user_config_file:
        logger.info('Loading opencue config from %s', user_config_file)
        with open(user_config_file, encoding="utf-8") as file_object:
            config.update(yaml.load(file_object, Loader=yaml.SafeLoader))

    return config



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module for communicating with the Cuebot server(s)."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import object
from random import shuffle
import abc
import time
import atexit
import logging
import os
import platform

import grpc

from opencue_proto import comment_pb2
from opencue_proto import comment_pb2_grpc
from opencue_proto import criterion_pb2
from opencue_proto import cue_pb2
from opencue_proto import cue_pb2_grpc
from opencue_proto import department_pb2
from opencue_proto import department_pb2_grpc
from opencue_proto import depend_pb2
from opencue_proto import depend_pb2_grpc
from opencue_proto import facility_pb2
from opencue_proto import facility_pb2_grpc
from opencue_proto import filter_pb2
from opencue_proto import filter_pb2_grpc
from opencue_proto import host_pb2
from opencue_proto import host_pb2_grpc
from opencue_proto import job_pb2
from opencue_proto import job_pb2_grpc
from opencue_proto import limit_pb2
from opencue_proto import limit_pb2_grpc
from opencue_proto import renderPartition_pb2
from opencue_proto import renderPartition_pb2_grpc
from opencue_proto import service_pb2
from opencue_proto import service_pb2_grpc
from opencue_proto import show_pb2
from opencue_proto import show_pb2_grpc
from opencue_proto import subscription_pb2
from opencue_proto import subscription_pb2_grpc
from opencue_proto import task_pb2
from opencue_proto import task_pb2_grpc
from opencue.exception import ConnectionException
from opencue.exception import CueException
import opencue.config


__all__ = ["Cuebot"]

logger = logging.getLogger("opencue")


DEFAULT_MAX_MESSAGE_BYTES = 1024 ** 2 * 10
DEFAULT_GRPC_PORT = 8443

if platform.system() != 'Darwin':
    # Avoid spamming users with epoll fork warning messages
    os.environ["GRPC_POLL_STRATEGY"] = "epoll1"

class Cuebot(object):
    """Used to manage the connection to the Cuebot.  Normally the connection
       to the Cuebot is made automatically as needed so you don't have to explicitly
       call Cuebot.connect().

       If you need to change the host(s) in which the library is connecting to,
       you have a couple options.  You can set it programmatically with
       Cuebot.setHosts or set the CUEBOT_HOSTS environment variable
       to a comma delimited list of host names."""
    RpcChannel = None
    Hosts = []
    Stubs = {}
    Config = opencue.config.load_config_from_file()
    Timeout = Config.get('cuebot.timeout', 10000)

    PROTO_MAP = {
        'action': filter_pb2,
        'allocation': facility_pb2,
        'comment': comment_pb2,
        'criterion': criterion_pb2,
        'cue': cue_pb2,
        'department': department_pb2,
        'depend': depend_pb2,
        'facility': facility_pb2,
        'filter': filter_pb2,
        'frame': job_pb2,
        'group': job_pb2,
        'host': host_pb2,
        'job': job_pb2,
        'layer': job_pb2,
        'limit': limit_pb2,
        'matcher': filter_pb2,
        'owner': host_pb2,
        'proc': host_pb2,
        'renderPartition': renderPartition_pb2,
        'service': service_pb2,
        'show': show_pb2,
        'subscription': subscription_pb2,
        'task': task_pb2
    }

    SERVICE_MAP = {
        'action': filter_pb2_grpc.ActionInterfaceStub,
        'allocation': facility_pb2_grpc.AllocationInterfaceStub,
        'comment': comment_pb2_grpc.CommentInterfaceStub,
        'cue': cue_pb2_grpc.CueInterfaceStub,
        'depend': depend_pb2_grpc.DependInterfaceStub,
        'department': department_pb2_grpc.DepartmentInterfaceStub,
        'facility': facility_pb2_grpc.FacilityInterfaceStub,
        'filter': filter_pb2_grpc.FilterInterfaceStub,
        'frame': job_pb2_grpc.FrameInterfaceStub,
        'group': job_pb2_grpc.GroupInterfaceStub,
        'host': host_pb2_grpc.HostInterfaceStub,
        'job': job_pb2_grpc.JobInterfaceStub,
        'layer': job_pb2_grpc.LayerInterfaceStub,
        'limit': limit_pb2_grpc.LimitInterfaceStub,
        'matcher': filter_pb2_grpc.MatcherInterfaceStub,
        'owner': host_pb2_grpc.OwnerInterfaceStub,
        'proc': host_pb2_grpc.ProcInterfaceStub,
        'renderPartition': renderPartition_pb2_grpc.RenderPartitionInterfaceStub,
        'service': service_pb2_grpc.ServiceInterfaceStub,
        'serviceOverride': service_pb2_grpc.ServiceOverrideInterfaceStub,
        'show': show_pb2_grpc.ShowInterfaceStub,
        'subscription': subscription_pb2_grpc.SubscriptionInterfaceStub,
        'task': task_pb2_grpc.TaskInterfaceStub
    }

    @staticmethod
    def init(config=None):
        """Main init method for setting up the Cuebot object.
        Sets the communication channel and hosts.

        :type config: dict
        :param config: config dictionary, this will override the config read from disk
        """
        hosts_env = os.getenv("CUEBOT_HOSTS")

        if config:
            Cuebot.Config = config
            Cuebot.Timeout = config.get('cuebot.timeout', Cuebot.Timeout)
        if hosts_env:
            Cuebot.setHosts(hosts_env.split(","))
        else:
            facility = os.getenv("CUEBOT_FACILITY", Cuebot.Config.get("cuebot.facility_default"))
            Cuebot.setHostWithFacility(facility)
        if Cuebot.Hosts is None:
            raise CueException('Cuebot host not set. Please ensure CUEBOT_HOSTS is set ' +
                               'or a facility_default host is set in the yaml pycue config.')

    @staticmethod
    def setChannel():
        """Sets the gRPC channel connection"""
        # gRPC must specify a single host. Randomize host list to balance load across cuebots.
        hosts = list(Cuebot.Hosts)
        shuffle(hosts)
        maxMessageBytes = Cuebot.Config.get('cuebot.max_message_bytes', DEFAULT_MAX_MESSAGE_BYTES)

        # create interceptors
        interceptors = (
            RetryOnRpcErrorClientInterceptor(
                max_attempts=4,
                sleeping_policy=ExponentialBackoff(init_backoff_ms=100,
                                                   max_backoff_ms=1600,
                                                   multiplier=2),
                status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
            ),
        )

        connectStr = "Not Defined"
        for host in hosts:
            if ':' in host:
                connectStr = host
            else:
                connectStr = '%s:%s' % (
                    host, Cuebot.Config.get('cuebot.grpc_port', DEFAULT_GRPC_PORT))
            # pylint: disable=logging-not-lazy
            logger.debug('connecting to gRPC at %s' % connectStr)
            # pylint: enable=logging-not-lazy
            # TODO(bcipriano) Configure gRPC TLS. (Issue #150)
            try:
                Cuebot.RpcChannel = grpc.intercept_channel(
                    grpc.insecure_channel(connectStr, options=[
                        ('grpc.max_send_message_length', maxMessageBytes),
                        ('grpc.max_receive_message_length', maxMessageBytes)]),
                    *interceptors)
                # Test the connection
                Cuebot.getStub('cue').GetSystemStats(
                    cue_pb2.CueGetSystemStatsRequest(), timeout=Cuebot.Timeout)
            # pylint: disable=broad-except
            except Exception:
                logger.warning('Could not establish grpc channel with %s', connectStr)
                continue
            atexit.register(Cuebot.closeChannel)
            return None
        raise ConnectionException('No grpc connection could be established. ' +
                                  'Please check configured cuebot hosts: ' + connectStr)

    @staticmethod
    def closeChannel():
        """Close the gRPC channel, delete it and reset it to None."""
        if Cuebot and Cuebot.RpcChannel is not None:
            Cuebot.RpcChannel.close()
            del Cuebot.RpcChannel
            Cuebot.RpcChannel = None

    @staticmethod
    def resetChannel():
        """Close and reopen the gRPC channel."""
        Cuebot.closeChannel()
        Cuebot.setChannel()

    @staticmethod
    def setHostWithFacility(facility):
        """Sets hosts to connect to based on the provided facility.
        If an unknown facility is provided, it will fall back to the one listed
        in cuebot.facility_default

        :type  facility: str
        :param facility: a facility named in the config file"""
        if facility not in list(Cuebot.Config.get("cuebot.facility").keys()):
            default = Cuebot.Config.get("cuebot.facility_default")
            logger.warning("The facility '%s' does not exist, defaulting to %s", facility, default)
            facility = default
        logger.debug("setting facility to: %s", facility)
        hosts = Cuebot.Config.get("cuebot.facility")[facility]
        Cuebot.setHosts(hosts)

    @staticmethod
    def setHosts(hosts):
        """Sets the cuebot host names to connect to.

        :param hosts: a list of hosts or a host
        :type hosts: list<str> or str"""
        if isinstance(hosts, str):
            hosts = [hosts]
        logger.debug("setting new server hosts to: %s", hosts)
        Cuebot.Hosts = hosts
        Cuebot.resetChannel()

    @staticmethod
    def setTimeout(timeout):
        """Sets the default network timeout.

        :param timeout: The network connection timeout in millis.
        :type timeout: int
        """
        logger.debug("setting new server timeout to: %d", timeout)
        Cuebot.Timeout = timeout

    @classmethod
    def getProto(cls, name):
        """Returns a proto class for the given name."""
        proto = cls.PROTO_MAP.get(name)
        if proto is None:
            raise ValueError("Could not find proto for {}.".format(name))
        return proto

    @classmethod
    def getService(cls, name):
        """Returns the service for the given name."""
        service = cls.SERVICE_MAP.get(name)
        if service is None:
            raise ValueError("Could not find stub interface for {}.".format(name))
        return service

    @classmethod
    def getStub(cls, name):
        """Get the matching stub from the SERVICE_MAP.
        Reuse an existing one if possible.

        :param name: name of stub key for SERVICE_MAP
        :type name: str"""
        if Cuebot.RpcChannel is None:
            cls.init()

        service = cls.getService(name)
        return service(Cuebot.RpcChannel)

    @staticmethod
    def getConfig():
        """Gets the Cuebot config object, originally read in from the config file on disk."""
        return Cuebot.Config


# Python 2/3 compatible implementation of ABC
ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})


class SleepingPolicy(ABC):
    """
    Implement policy for sleeping between API retries
    """
    @abc.abstractmethod
    def sleep(self, attempt):
        """
        How long to sleep in milliseconds.
        :param attempt: the number of attempt (starting from zero)
        """
        assert attempt >= 0


class ExponentialBackoff(SleepingPolicy):
    """
    Implement policy that will increase retry period by exponentially in every try
    """
    def __init__(self,
                 init_backoff_ms,
                 max_backoff_ms,
                 multiplier=2):
        """
        inputs in ms
        """
        self._init_backoff = init_backoff_ms
        self._max_backoff = max_backoff_ms
        self._multiplier = multiplier

    def sleep(self, attempt):
        sleep_time_ms = min(
            self._init_backoff * self._multiplier ** attempt,
            self._max_backoff
        )
        time.sleep(sleep_time_ms / 1000.0)


class RetryOnRpcErrorClientInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.StreamUnaryClientInterceptor
):
    """
    Implement Client/Stream interceptors for GRPC channels to retry
    calls that failed with retry-able states. This is required for
    handling server interruptions that are not automatically handled
    by grpc.insecure_channel
    """
    def __init__(self,
                 max_attempts,
                 sleeping_policy,
                 status_for_retry=None):
        self._max_attempts = max_attempts
        self._sleeping_policy = sleeping_policy
        self._retry_statuses = status_for_retry

    # pylint: disable=inconsistent-return-statements
    def _intercept_call(self, continuation, client_call_details,
                        request_or_iterator):
        for attempt in range(self._max_attempts):
            try:
                return continuation(client_call_details,
                                    request_or_iterator)
            except grpc.RpcError as response:
                # Return if it was last attempt
                if attempt == (self._max_attempts - 1):
                    return response

                # If status code is not in retryable status codes
                # pylint: disable=no-member
                if self._retry_statuses \
                        and hasattr(response, 'code') \
                        and response.code() \
                        not in self._retry_statuses:
                    return response

                self._sleeping_policy.sleep(attempt)

    def intercept_unary_unary(self, continuation, client_call_details,
                              request):
        return self._intercept_call(continuation, client_call_details,
                                    request)

    def intercept_stream_unary(
            self, continuation, client_call_details, request_iterator
    ):
        return self._intercept_call(continuation, client_call_details,
                                    request_iterator)



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Helper class for representing a frame range.

It supports a complex syntax implementing features such as comma-separated frame ranges,
stepped frame ranges and more. See the FrameRange class for more detail.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from builtins import str
from builtins import range
from builtins import object
import re
from collections import OrderedDict


class FrameRange(object):
    """Represents a sequence of frame numbers."""

    SINGLE_FRAME_PATTERN = re.compile(r'^(-?)\d+$')
    SIMPLE_FRAME_RANGE_PATTERN = re.compile(r'^(?P<sf>(-?)\d+)-(?P<ef>(-?)\d+)$')
    STEP_PATTERN = re.compile(
        r'^(?P<sf>(-?)\d+)-(?P<ef>(-?)\d+)(?P<stepSep>[xy])(?P<step>(-?)\d+)$')
    INTERLEAVE_PATTERN = re.compile(r'^(?P<sf>(-?)\d+)-(?P<ef>(-?)\d+):(?P<step>(-?)\d+)$')

    def __init__(self, frameRange):
        """
        Construct a FrameRange object by parsing a spec.

        FrameSet("1-10x3");
        FrameSet("1-10y3"); // inverted step
        FrameSet("10-1x-1");
        FrameSet("1"); // same as "1-1x1"
        FrameSet("1-10:5"); // interleave of 5

        A valid spec consists of:

        An inTime.
        An optional hyphen and outTime.
        An optional x or y and stepSize.
        Or an optional : and interleaveSize.
        If outTime is less than inTime, stepSize must be negative.

        A stepSize of 0 produces an empty FrameRange.

        A stepSize cannot be combined with a interleaveSize.

        A stepSize designated with y creates an inverted step. Frames that would be included
        with an x step are excluded.

        Example: 1-10y3 == 2, 3, 5, 6, 8, 9.

        An interleaveSize alters the order of frames when iterating over the FrameRange. The
        iterator will first produce the list of frames from inTime to outTime with a stepSize
        equal to interleaveSize. The interleaveSize is then divided in half, producing another
        set of frames unique from the first set. This process is repeated until interleaveSize
        reaches 1.

        Example: 1-10:5 == 1, 6, 2, 4, 8, 10, 3, 5, 7, 9.
        """
        self.frameList = self.parseFrameRange(frameRange)

    def __str__(self):
        # TODO(bcipriano) Make this smarter, group frame ranges and by step. (Issue #83)
        return ','.join([str(frame) for frame in self.frameList])

    def __getitem__(self, key):
        return self.frameList[key]

    def __len__(self):
        return self.size()

    def size(self):
        """Gets the number of frames contained in this sequence."""
        return len(self.frameList)

    def get(self, idx):
        """Gets an individual entry in the sequence, by numerical position."""
        return self.frameList[idx]

    def index(self, idx):
        """Query index of frame number in frame set.

        Returns:
            int, index of frame. -1 if frame set does not contain frame.
        """
        try:
            return self.frameList.index(idx)
        except ValueError:
            return -1

    def getAll(self):
        """Gets the full numerical sequence."""
        return self.frameList

    def normalize(self):
        """Sorts and deduplicates the sequence."""
        self.frameList = list(set(self.frameList))
        self.frameList.sort()

    @classmethod
    def parseFrameRange(cls, frameRange):
        """
        Parse a string representation into a numerical sequence.

        :type frameRange: str
        :param frameRange: String representation of the frame range.
        :rtype: FrameRange
        :return: FrameRange representing the numerical sequence.
        """
        singleFrameMatcher = re.match(cls.SINGLE_FRAME_PATTERN, frameRange)
        if singleFrameMatcher:
            return [int(frameRange)]

        simpleRangeMatcher = re.match(cls.SIMPLE_FRAME_RANGE_PATTERN, frameRange)
        if simpleRangeMatcher:
            startFrame = int(simpleRangeMatcher.group('sf'))
            endFrame = int(simpleRangeMatcher.group('ef'))
            return cls.__getIntRange(startFrame, endFrame, (1 if endFrame >= startFrame else -1))

        rangeWithStepMatcher = re.match(cls.STEP_PATTERN, frameRange)
        if rangeWithStepMatcher:
            startFrame = int(rangeWithStepMatcher.group('sf'))
            endFrame = int(rangeWithStepMatcher.group('ef'))
            step = int(rangeWithStepMatcher.group('step'))
            stepSep = rangeWithStepMatcher.group('stepSep')
            return cls.__getSteppedRange(startFrame, endFrame, step, stepSep == 'y')

        rangeWithInterleaveMatcher = re.match(cls.INTERLEAVE_PATTERN, frameRange)
        if rangeWithInterleaveMatcher:
            startFrame = int(rangeWithInterleaveMatcher.group('sf'))
            endFrame = int(rangeWithInterleaveMatcher.group('ef'))
            step = int(rangeWithInterleaveMatcher.group('step'))
            return cls.__getInterleavedRange(startFrame, endFrame, step)

        raise ValueError('unrecognized frame range syntax ' + frameRange)

    @staticmethod
    def __getIntRange(start, end, step):
        return list(range(start, end+(step // abs(step)), step))

    @classmethod
    def __getSteppedRange(cls, start, end, step, inverseStep):
        cls.__validateStepSign(start, end, step)
        steppedRange = cls.__getIntRange(start, end, step)
        if inverseStep:
            fullRange = cls.__getIntRange(start, end, (-1 if step < 0 else 1))
            return [frame for frame in fullRange if frame not in steppedRange]
        return steppedRange

    @classmethod
    def __getInterleavedRange(cls, start, end, step):
        cls.__validateStepSign(start, end, step)
        interleavedFrames = OrderedDict()
        incrValue = step // abs(step)
        while abs(step) > 0:
            interleavedFrames.update(
                [(frame, None) for frame in cls.__getIntRange(start, end, step)])
            start += incrValue
            step = int(step / 2.0)
        return list(interleavedFrames.keys())

    @staticmethod
    def __validateStepSign(start, end, step):
        if step > 1 and end < start:
            raise ValueError(
                'end frame may not be less than start frame when using a positive step')
        if step == 0:
            raise ValueError('step cannot be zero')
        if step < 0 and end >= start:
            raise ValueError(
                'end frame may not be greater than start frame when using a negative step')



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module for classes related to job layers."""

import enum
import getpass
import os
import platform

from opencue_proto import job_pb2
import opencue.api
from opencue.cuebot import Cuebot
import opencue.search
import opencue.wrappers.depend
import opencue.wrappers.frame
import opencue.wrappers.limit


class Layer(object):
    """This class contains the grpc implementation related to a Layer."""

    class LayerType(enum.IntEnum):
        """Represents the type of layer."""
        PRE = job_pb2.PRE
        POST = job_pb2.POST
        RENDER = job_pb2.RENDER
        UTIL = job_pb2.UTIL

    class Order(enum.IntEnum):
        """Represents the order of a layer."""
        FIRST = job_pb2.FIRST
        LAST = job_pb2.LAST
        REVERSE = job_pb2.REVERSE

    def __init__(self, layer=None):
        self.data = layer
        self.stub = Cuebot.getStub('layer')

    def kill(self, username=None, pid=None, host_kill=None, reason=None):
        """Kills the entire layer."""
        username = username if username else getpass.getuser()
        pid = pid if pid else os.getpid()
        host_kill = host_kill if host_kill else platform.uname()[1]
        return self.stub.KillFrames(job_pb2.LayerKillFramesRequest(layer=self.data,
                                                                   username=username,
                                                                   pid=str(pid),
                                                                   host_kill=host_kill,
                                                                   reason=reason),
                                    timeout=Cuebot.Timeout)

    def eat(self):
        """Eats the entire layer."""
        return self.stub.EatFrames(job_pb2.LayerEatFramesRequest(layer=self.data),
                                   timeout=Cuebot.Timeout)

    def retry(self):
        """Retries the entire layer."""
        return self.stub.RetryFrames(job_pb2.LayerRetryFramesRequest(layer=self.data),
                                     timeout=Cuebot.Timeout)

    def markdone(self):
        """Drops any dependency that requires this layer or requires any frame in the layer."""
        return self.stub.MarkdoneFrames(job_pb2.LayerMarkdoneFramesRequest(layer=self.data),
                                        timeout=Cuebot.Timeout)

    def addLimit(self, limit_id):
        """Adds a limit to the current layer."""
        return self.stub.AddLimit(job_pb2.LayerAddLimitRequest(layer=self.data, limit_id=limit_id),
                                  timeout=Cuebot.Timeout)

    def dropLimit(self, limit_id):
        """Removes a limit on the current layer."""
        return self.stub.DropLimit(
            job_pb2.LayerDropLimitRequest(layer=self.data, limit_id=limit_id),
            timeout=Cuebot.Timeout)

    def enableMemoryOptimizer(self, value):
        """Enables or disables the memory optimizer.

        :type  value: bool
        :param value: whether memory optimizer is enabled
        """
        return self.stub.EnableMemoryOptimizer(job_pb2.LayerEnableMemoryOptimizerRequest(
            layer=self.data, value=value),
            timeout=Cuebot.Timeout)

    def getFrames(self, **options):
        """Returns a list of up to 1000 frames from within the layer.

        :type  options: dict
        :param options: FrameSearch options
        :rtype:  list<opencue.wrappers.frame.Frame>
        :return: sequence of matching frames
        """
        criteria = opencue.search.FrameSearch.criteriaFromOptions(**options)
        response = self.stub.GetFrames(job_pb2.LayerGetFramesRequest(layer=self.data, s=criteria),
                                       timeout=Cuebot.Timeout)
        return [opencue.wrappers.frame.Frame(frameData) for frameData in response.frames.frames]

    def getOutputPaths(self):
        """Return the output paths for this layer.

        :rtype: list<str>
        :return: list of output paths
        """
        return self.stub.GetOutputPaths(job_pb2.LayerGetOutputPathsRequest(layer=self.data),
                                        timeout=Cuebot.Timeout).output_paths

    def setTags(self, tags):
        """Sets the layer tags.

        :type  tags: list<str>
        :param tags: layer tags
        """
        return self.stub.SetTags(job_pb2.LayerSetTagsRequest(layer=self.data, tags=tags),
                                 timeout=Cuebot.Timeout)

    def setMaxCores(self, cores):
        """Sets the maximum number of cores that this layer requires.

        :type  cores: float
        :param cores: Core units, 100 reserves 1 core
        """
        return self.stub.SetMaxCores(
            job_pb2.LayerSetMaxCoresRequest(layer=self.data, cores=cores/100.0),
            timeout=Cuebot.Timeout)

    def setMinCores(self, cores):
        """Sets the minimum number of cores that this layer requires.

        Use 100 to reserve 1 core.

        :type  cores: int
        :param cores: core units, 100 reserves 1 core
        """
        return self.stub.SetMinCores(
            job_pb2.LayerSetMinCoresRequest(layer=self.data, cores=cores/100.0),
            timeout=Cuebot.Timeout)

    def setMaxGpus(self, max_gpus):
        """Sets the maximum number of gpus that this layer requires.
        :type  max_gpus: int
        :param max_gpus: gpu cores"""
        return self.stub.SetMaxGpus(
            job_pb2.LayerSetMaxGpusRequest(layer=self.data, max_gpus=max_gpus),
            timeout=Cuebot.Timeout)

    def setMinGpus(self, min_gpus):
        """Sets the minimum number of gpus that this layer requires.
        :type  min_gpus: int
        :param min_gpus: gou cores"""
        return self.stub.SetMinGpus(
            job_pb2.LayerSetMinGpusRequest(layer=self.data, min_gpus=min_gpus),
            timeout=Cuebot.Timeout)

    def setMinGpuMemory(self, gpu_memory):
        """Sets the minimum number of gpu memory that this layer requires.

        :type  gpu_memory: int
        :param gpu_memory: gpu_memory value
        """
        return self.stub.SetMinGpuMemory(
            job_pb2.LayerSetMinGpuMemoryRequest(layer=self.data, gpu_memory=gpu_memory),
            timeout=Cuebot.Timeout)

    def setMinMemory(self, memory):
        """Sets the minimum amount of memory that this layer requires.

        :type  memory: int
        :param memory: Minimum Kb memory reserved by each frame
        """
        return self.stub.SetMinMemory(
            job_pb2.LayerSetMinMemoryRequest(layer=self.data, memory=memory),
            timeout=Cuebot.Timeout)

    def setThreadable(self, threadable):
        """Sets the threadable field.

        :type  threadable: bool
        :param threadable: boolean to enable/disable threadable
        """
        return self.stub.SetThreadable(job_pb2.LayerSetThreadableRequest(
            layer=self.data, threadable=threadable),
            timeout=Cuebot.Timeout)

    def setTimeout(self, timeout):
        """Set time out to the value.
        :type timeout: int
        :param timeout: value for timeout in minutes"""
        return self.stub.SetTimeout(job_pb2.LayerSetTimeoutRequest(
            layer=self.data, timeout=timeout),
            timeout=Cuebot.Timeout)

    def setTimeoutLLU(self, timeout_llu):
        """Set LLU time out to the value.
        :type timeout: int
        :param timeout: value for timeout in minutes"""
        return self.stub.SetTimeoutLLU(job_pb2.LayerSetTimeoutLLURequest(
            layer=self.data, timeout_llu=timeout_llu),
            timeout=Cuebot.Timeout)

    def addRenderPartition(self, hostname, threads, max_cores, max_mem, max_gpu_memory, max_gpus):
        """Adds a render partition to the layer.

        :type  hostname: str
        :param hostname: hostname of the partition
        :type  threads: int
        :param threads: number of threads of the partition
        :type  max_cores: int
        :param max_cores: max cores enabled for the partition
        :type  max_mem: int
        :param max_mem: amount of memory reserved for the partition
        :type  max_gpu_memory: int
        :param max_gpu_memory: max gpu memory enabled for the partition
        :type  max_gpus: int
        :param max_gpus: max gpus enabled for the partition
        """
        self.stub.AddRenderPartition(
            job_pb2.LayerAddRenderPartitionRequest(layer=self.data,
                                                   host=hostname,
                                                   threads=threads,
                                                   max_cores=max_cores,
                                                   max_memory=max_mem,
                                                   max_gpu_memory=max_gpu_memory,
                                                   username=os.getenv("USER", "unknown"),
                                                   max_gpus=max_gpus))

    def getWhatDependsOnThis(self):
        """Gets a list of dependencies that depend directly on this layer.

        :rtype:  list<opencue.wrappers.depend.Depend>
        :return: list of dependencies that depend directly on this layer
        """
        response = self.stub.GetWhatDependsOnThis(
            job_pb2.LayerGetWhatDependsOnThisRequest(layer=self.data),
            timeout=Cuebot.Timeout)
        dependSeq = response.depends
        return [opencue.wrappers.depend.Depend(dep) for dep in dependSeq.depends]

    def getWhatThisDependsOn(self):
        """Returns a list of dependencies that this layer depends on.

        :rtype:  list<opencue.wrappers.depend.Depend>
        :return: list of dependences that this layer depends on
        """
        response = self.stub.GetWhatThisDependsOn(
            job_pb2.LayerGetWhatThisDependsOnRequest(layer=self.data),
            timeout=Cuebot.Timeout)
        dependSeq = response.depends
        return [opencue.wrappers.depend.Depend(dep) for dep in dependSeq.depends]

    def createDependencyOnJob(self, job):
        """Creates and returns a layer-on-job dependency.

        :type  job: opencue.wrappers.job.Job
        :param job: the job you want this job to depend on
        :rtype:  opencue.wrappers.depend.Depend
        :return: the new dependency
        """
        response = self.stub.CreateDependencyOnJob(
            job_pb2.LayerCreateDependOnJobRequest(layer=self.data, job=job.data),
            timeout=Cuebot.Timeout)
        return opencue.wrappers.depend.Depend(response.depend)

    def createDependencyOnLayer(self, layer):
        """Creates and returns a layer-on-layer dependency.

        :type  layer: opencue.wrappers.layer.Layer
        :param layer: the layer you want this layer to depend on
        :rtype:  opencue.wrappers.depend.Depend
        :return: the new dependency
        """
        response = self.stub.CreateDependencyOnLayer(
            job_pb2.LayerCreateDependOnLayerRequest(layer=self.data, depend_on_layer=layer.data),
            timeout=Cuebot.Timeout)
        return opencue.wrappers.depend.Depend(response.depend)

    def createDependencyOnFrame(self, frame):
        """Creates and returns a layer-on-frame dependency.

        :type  frame: opencue.wrappers.frame.Frame
        :param frame: the frame you want this layer to depend on
        :rtype:  opencue.wrappers.depend.Depend
        :return: the new dependency
        """
        response = self.stub.CreateDependencyOnFrame(
            job_pb2.LayerCreateDependOnFrameRequest(layer=self.data, frame=frame.data),
            timeout=Cuebot.Timeout)
        return opencue.wrappers.depend.Depend(response.depend)

    def createFrameByFrameDependency(self, layer):
        """Creates and returns a frame-by-frame layer dependency.

        :param layer: the layer you want this layer to depend on
        :type  layer: opencue.wrappers.layer.Layer
        :rtype:  opencue.wrappers.depend.Depend
        :return: the new dependency
        """
        # anyframe is hard coded right now, this option should be moved
        # to LayerOnLayer for better efficiency.
        response = self.stub.CreateFrameByFrameDependency(
            job_pb2.LayerCreateFrameByFrameDependRequest(
                layer=self.data, depend_layer=layer.data, any_frame=False),
            timeout=Cuebot.Timeout)
        return opencue.wrappers.depend.Depend(response.depend)

    # TODO(gregdenton) Determine if this is needed. (Issue #71)
    # def unbookProcs(self, subs, number, kill=False):
    #     """Unbook procs off layer from specified subscriptions
    #     :type  subs: list<Subscription>
    #     :param subs: the subscriptions to unbook from
    #     :type  number: int
    #     :param number: the number of virtual procs to unbook
    #     :type  kill: bool
    #     :param kill: wheather or not to kill the frames as well"""
    #     self.proxy.unbookProcs([a.proxy for a in subs], number, kill)

    def registerOutputPath(self, outputPath):
        """Registers an output path for the layer.

        Output paths are included in OpenCue alert emails.

        :type  outputPath: str
        :param outputPath: output path to register
        """
        self.stub.RegisterOutputPath(
            job_pb2.LayerRegisterOutputPathRequest(layer=self.data, spec=outputPath),
            timeout=Cuebot.Timeout)

    def reorderFrames(self, frameRange, order):
        """Reorders the specified frame range on this layer.

        :type  frameRange: string
        :param frameRange: the frame range to reorder
        :type  order: opencue.wrapper.layer.Layer.Order
        :param order: First, Last or Reverse
        """
        self.stub.ReorderFrames(
            job_pb2.LayerReorderFramesRequest(layer=self.data, range=frameRange, order=order),
            timeout=Cuebot.Timeout)

    def staggerFrames(self, frameRange, stagger):
        """Staggers the specified frame range on this layer.

        :type  frameRange: string
        :param frameRange: the frame range to stagger
        :type  stagger: int
        :param stagger: the amount to stagger by
        """
        self.stub.StaggerFrames(
            job_pb2.LayerStaggerFramesRequest(layer=self.data, range=frameRange, stagger=stagger),
            timeout=Cuebot.Timeout)

    def getLimitDetails(self):
        """Returns the Limit objects for the given layer.

        :rtype:  list<opencue.wrappers.limit.Limit>
        :return: list of limits on this layer
        """
        return [opencue.wrappers.limit.Limit(limit) for limit in self.stub.GetLimits(
            job_pb2.LayerGetLimitsRequest(layer=self.data), timeout=Cuebot.Timeout).limits]

    def id(self):
        """Returns the id of the layer.

        :rtype:  str
        :return: layer id
        """
        return self.data.id

    def name(self):
        """Returns the name of the layer.

        :rtype:  str
        :return: layer name
        """
        return self.data.name

    def range(self):
        """Returns the frame range of the layer.

        :rtype:  str
        :return: layer frame range
        """
        return self.data.range

    def chunkSize(self):
        """Returns the number of frames per task.

        :rtype:  int
        :return: layer chunk size
        """
        return self.data.chunk_size

    def tags(self):
        """Returns the tags applied to the layer.

        TODO: Document syntax

        :rtype:  str
        :return: layer tags"""
        return self.data.tags

    def dispatchOrder(self):
        """Returns the layer dispatch order.

        :rtype:  int
        :return: layer dispatch order
        """
        return self.data.dispatch_order

    def coresReserved(self):
        """Returns the number of cores reserved on this layer.

        :rtype: float
        :return: cores reserved
        """
        return self.data.layer_stats.reserved_cores

    def gpusReserved(self):
        """Returns the number of gpus reserved on this layer
        :rtype: float
        :return: gpus reserved"""
        return self.data.layer_stats.reserved_gpus

    def minCores(self):
        """Returns the minimum number of cores that frames in this layer require.

        :rtype:  int
        :return: minimum number of cores required
        """
        return self.data.min_cores

    def minGpus(self):
        """Returns the minimum number of gpus that frames in this layer require
        :rtype:  int
        :return: Minimum number of gpus required"""
        return self.data.min_gpus

    def minMemory(self):
        """Returns the minimum amount of memory that frames in this layer require.

        :rtype:  int
        :return: minimum kB of memory required by frames in this layer
        """
        return self.data.min_memory

    def limits(self):
        """Returns the limit names for this layer.

        :rtype: list<str>
        :return: names of the limits on this layer
        """
        return self.data.limits

    def maxRss(self):
        """Returns the highest amount of memory that any frame in this layer used.

        Value is within 5% of the actual highest frame.

        :rtype:  long
        :return: most memory used by any frame in this layer in kB"""
        return self.data.layer_stats.max_rss

    def type(self):
        """Returns the type of layer.

        Ex: Pre, Post, Render

        :rtype:  job_pb2.LayerType
        :return: type of layer
        """
        return self.data.type

    def totalFrames(self):
        """Returns the total number of frames in the layer.

        :rtype:  int
        :return: total number of frames
        """
        return self.data.layer_stats.total_frames

    def dependFrames(self):
        """Returns the total number of dependent frames in the layer.

        :rtype:  int
        :return: total number of dependent frames
        """
        return self.data.layer_stats.depend_frames

    def succeededFrames(self):
        """Returns the total number of succeeded frames in the layer.

        :rtype:  int
        :return: total number of succeeded frames
        """
        return self.data.layer_stats.succeeded_frames

    def runningFrames(self):
        """Returns the total number of running frames in the layer.

        :rtype:  int
        :return: total number of running frames
        """
        return self.data.layer_stats.running_frames

    def deadFrames(self):
        """Returns the total number of dead frames in the layer.

        :rtype:  int
        :return: total number of dead frames
        """
        return self.data.layer_stats.dead_frames

    def waitingFrames(self):
        """Returns the total number of waiting frames in the layer.

        :rtype:  int
        :return: total number of waiting frames
        """
        return self.data.layer_stats.waiting_frames

    def eatenFrames(self):
        """Returns the total number of eaten frames in the layer.

        :rtype:  int
        :return: total number of eaten frames
        """
        return self.data.layer_stats.eaten_frames

    def pendingFrames(self):
        """Returns the total number of pending (dependent and waiting) frames in the layer.

        :rtype:  int
        :return: total number of pending (dependent and waiting) frames
        """
        return self.data.layer_stats.pending_frames

    def percentCompleted(self):
        """Returns the percent that the layer's frames are completed.

        :rtype:  float
        :return: percentage of frame completion
        """
        try:
            return self.data.layer_stats.succeeded_frames / \
                   float(self.data.layer_stats.total_frames) * 100.0
        except ZeroDivisionError:
            return 0

    def avgFrameTimeSeconds(self):
        """Returns the average frame completion time in seconds.

        :rtype:  int
        :return: average frame completion time in seconds
        """
        return self.data.layer_stats.avg_frame_sec

    def avgCoreSeconds(self):
        """Returns the average core time used in seconds.

        :rtype:  int
        :return: average core time in seconds
        """
        return self.data.layer_stats.avg_core_sec

    def coreSecondsRemaining(self):
        """Returns the estimated core time that is remaining to complete all waiting frames.

        :rtype:  int
        :return: the number of seconds of estimated core time remaining
        """
        return self.data.layer_stats.remaining_core_sec

    def parent(self):
        """Gets the parent of the layer; its job.

        :rtype: opencue.wrappers.job.Job
        :return: the layer's parent job
        """
        return opencue.api.getJob(self.data.parent_id)

    def services(self):
        """Returns list of services applied to this layer
        :rtype: opencue.wrappers.service.Service
        :return: the layer's services
        """
        return [opencue.api.getService(service) for service in self.data.services]
